#!/usr/bin/env python
#
#  iterative.py
"""
Functions for iteration, looping etc.

.. versionadded:: 1.4.0
"""
#
#  Copyright © 2018-2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#
#  chunks from https://stackoverflow.com/a/312464/3092681
#  Copyright © 2008 Ned Batchelder
#  Licensed under CC-BY-SA
#

# stdlib
import itertools
import textwrap
from operator import itemgetter
from typing import (
		Any,
		Callable,
		Iterable,
		Iterator,
		List,
		Optional,
		Sequence,
		Sized,
		Tuple,
		Type,
		TypeVar,
		Union,
		cast
		)

# 3rd party
from natsort import natsorted, ns
from typing_extensions import final

# this package
from domdf_python_tools.utils import magnitude

__all__ = [
		"chunks",
		"permutations",
		"split_len",
		"Len",
		"double_chain",
		"flatten",
		"make_tree",
		"natmin",
		"natmax",
		"groupfloats",
		"ranges_from_iterable",
		"extend",
		"extend_with",
		"extend_with_none",
		"count",
		"AnyNum",
		]

_T = TypeVar("_T")
AnyNum = TypeVar("AnyNum", float, complex)


def chunks(l: Sequence[_T], n: int) -> Iterator[Sequence[_T]]:
	"""
	Yield successive ``n``-sized chunks from ``l``.

	:param l: The objects to yield chunks from.
	:param n: The size of the chunks.

	:rtype:

	.. versionchanged:: 1.4.0 Moved from :mod:`domdf_python_tools.utils`
	"""

	for i in range(0, len(l), n):
		yield l[i:i + n]


def permutations(data: Iterable[_T], n: int = 2) -> List[Tuple[_T, ...]]:
	"""
	Return permutations containing ``n`` items from ``data`` without any reverse duplicates.

	If ``n`` is equal to or greater than the length of the data an empty list of returned.

	:param data:
	:param n:

	:rtype:

	.. versionchanged:: 1.4.0 Moved from :mod:`domdf_python_tools.utils`
	.. seealso:: :func:`itertools.permutations` and :func:`itertools.combinations`
	.. latex:clearpage::
	"""

	if n == 0:
		raise ValueError("'n' cannot be 0")

	perms = []
	for i in itertools.permutations(data, n):
		if i[::-1] not in perms:
			perms.append(i)

	return perms


def split_len(string: str, n: int) -> List[str]:
	"""
	Split ``string`` every ``n`` characters.

	:param string:
	:param n: The number of characters to split after

	:return: The split string

	.. versionchanged:: 1.4.0 Moved from :mod:`domdf_python_tools.utils`
	"""

	return [string[i:i + n] for i in range(0, len(string), n)]


def Len(obj: Sized, start: int = 0, step: int = 1) -> range:
	"""
	Shorthand for ``range(len(obj))``.

	Returns an object that produces a sequence of integers from ``start`` (inclusive)
	to :func:`len(obj) <len>` (exclusive) by ``step``.

	.. versionadded:: 0.4.7

	:param obj: The object to iterate over the length of.
	:param start: The start value of the range.
	:param step: The step of the range.

	:rtype:

	.. versionchanged:: 1.4.0 Moved from :mod:`domdf_python_tools.utils`
	"""

	return range(start, len(obj), step)


def double_chain(iterable: Iterable[Iterable[Iterable[_T]]]) -> Iterator[_T]:
	"""
	Flatten a list of lists of lists into a single list.

	Literally just:

	.. code-block:: python

		chain.from_iterable(chain.from_iterable(iterable))

	.. compound::

		Will convert

		.. code-block:: python

			[[(1, 2), (3, 4)], [(5, 6), (7, 8)]]

		to

		.. code-block:: python

			[1, 2, 3, 4, 5, 6, 7, 8]

	.. versionadded:: 0.4.7

	:param iterable: The iterable to chain.

	:rtype:

	.. versionchanged:: 1.4.0 Moved from :mod:`domdf_python_tools.utils`
	"""

	return itertools.chain.from_iterable(itertools.chain.from_iterable(iterable))


def flatten(iterable: Iterable[_T], primitives: Tuple[Type, ...] = (str, int, float)) -> Iterator[_T]:
	"""
	Flattens a mixed list of primitive types and iterables of those types into a single list,
	regardless of nesting.

	.. versionadded:: 1.4.0

	:param iterable:
	:param primitives: The primitive types to allow.
	"""  # noqa: D400

	for item in iterable:
		if isinstance(item, primitives):
			yield item
		elif isinstance(item, Iterable):
			yield from flatten(item)
		else:
			raise NotImplementedError


Branch = Union[Sequence[str], Sequence[Union[Sequence[str], Sequence]]]


def make_tree(tree: Branch) -> Iterator[str]:
	"""
	Returns the string representation of a mixed list of strings and lists of strings,
	similar to :manpage:`tree(1)`.

	.. versionadded:: 1.4.0

	:param tree:
	"""  # noqa: D400

	last_string = 0
	for idx, entry in enumerate(tree):
		if isinstance(entry, str):
			last_string = idx

	for idx, entry in enumerate(tree[:-1]):
		if isinstance(entry, str):
			if idx > last_string:
				yield f"│   {entry}"
			elif idx == last_string:
				yield f"└── {entry}"
			else:
				yield f"├── {entry}"

		elif isinstance(entry, Iterable):
			for line in make_tree(entry):
				if idx - 1 == last_string:
					yield textwrap.indent(line, "└── ")
				else:
					yield textwrap.indent(line, "│   ")

	if tree:
		if isinstance(tree[-1], str):
			yield f"└── {tree[-1]}"
		elif isinstance(tree[-1], Iterable):
			for line in make_tree(tree[-1]):
				yield textwrap.indent(line, "    ")


def natmin(seq: Iterable[_T], key: Optional[Callable[[Any], Any]] = None, alg: int = ns.DEFAULT) -> _T:
	"""
	Returns the minimum value from ``seq`` when sorted naturally.

	.. versionadded:: 1.8.0

	:param seq:
	:param key: A key used to determine how to sort each element of the iterable.
		It is **not** applied recursively.
		The callable should accept a single argument and return a single value.
	:param alg: This option is used to control which algorithm :mod:`natsort` uses when sorting.
	"""

	return natsorted(seq, key=key, alg=cast(ns, alg))[0]


def natmax(seq: Iterable[_T], key: Optional[Callable[[Any], Any]] = None, alg: int = ns.DEFAULT) -> _T:
	"""
	Returns the maximum value from ``seq`` when sorted naturally.

	.. versionadded:: 1.8.0

	:param seq:
	:param key: A key used to determine how to sort each element of the iterable.
		It is **not** applied recursively.
		The callable should accept a single argument and return a single value.
	:param alg: This option is used to control which algorithm :mod:`natsort` uses when sorting.
	"""

	return natsorted(seq, key=key, alg=cast(ns, alg))[-1]


_group = Tuple[float, ...]


def groupfloats(
		iterable: Iterable[float],
		step: float = 1,
		) -> Iterable[_group]:
	"""
	Returns an iterator over the discrete ranges of values in ``iterable``.

	For example:

	.. code-block:: python

		>>> list(groupfloats(
		... 	[170.0, 170.05, 170.1, 170.15, 171.05, 171.1, 171.15, 171.2],
		... 	step=0.05,
		... ))
		[(170.0, 170.05, 170.1, 170.15), (171.05, 171.1, 171.15, 171.2)]
		>>> list(groupfloats([1, 2, 3, 4, 5, 7, 8, 9, 10]))
		[(1, 2, 3, 4, 5), (7, 8, 9, 10)]

	.. versionadded:: 2.0.0

	:param iterable:
	:param step: The step between values in ``iterable``.

	:rtype:

	.. seealso::

		:func:`~.ranges_from_iterable`, which returns an iterator over the min and max values for each range.
	"""

	# Based on https://stackoverflow.com/a/4629241
	# By user97370
	# CC BY-SA 4.0

	modifier = 1 / 10**magnitude(step)

	a: float
	b: Iterable[_group]

	def key(pair):
		return (pair[1] * modifier) - ((pair[0] * modifier) * step)

	for a, b in itertools.groupby(enumerate(iterable), key=key):
		yield tuple(map(itemgetter(1), list(b)))


def ranges_from_iterable(iterable: Iterable[float], step: float = 1) -> Iterable[Tuple[float, float]]:
	"""
	Returns an iterator over the minimum and maximum values for each discrete ranges of values in ``iterable``.

	For example:

	.. code-block:: python

		>>> list(ranges_from_iterable([170.0, 170.05, 170.1, 170.15, 171.05, 171.1, 171.15, 171.2], step=0.05))
		[(170.0, 170.15), (171.05, 171.2)]
		>>> list(ranges_from_iterable([1, 2, 3, 4, 5, 7, 8, 9, 10]))
		[(1, 5), (7, 10)]

	:param iterable:
	:param step: The step between values in ``iterable``.
	"""

	for group in groupfloats(iterable, step):
		yield group[0], group[-1]


def extend(sequence: Iterable[_T], minsize: int) -> List[_T]:
	"""
	Extend ``sequence`` by repetition until it is at least as long as ``minsize``.

	.. versionadded:: 2.3.0

	:param sequence:
	:param minsize:

	:rtype:

	.. seealso:: :func:`~.extend_with` and :func:`~.extend_with_none`
	"""

	output = list(sequence)
	cycle = itertools.cycle(output)

	while len(output) < minsize:
		output.append(next(cycle))

	return output


def extend_with(sequence: Iterable[_T], minsize: int, with_: _T) -> List[_T]:
	r"""
	Extend ``sequence`` by adding ``with\_`` to the right hand end until it is at least as long as ``minsize``.

	.. versionadded:: 2.3.0

	:param sequence:
	:param minsize:
	:param with\_:

	:rtype:

	.. seealso:: :func:`~.extend` and :func:`~.extend_with_none`
	.. latex:clearpage::
	"""

	output = list(sequence)

	while len(output) < minsize:
		output.append(with_)

	return output


def extend_with_none(sequence: Iterable[_T], minsize: int) -> Sequence[Optional[_T]]:
	r"""
	Extend ``sequence`` by adding :py:obj:`None` to the right hand end until it is at least as long as ``minsize``.

	.. versionadded:: 2.3.0

	:param sequence:
	:param minsize:

	:rtype:

	.. seealso:: :func:`~.extend` and :func:`~.extend_with`
	"""

	output: Sequence[Optional[_T]] = list(sequence)
	filler: Sequence[Optional[_T]] = [None] * max(0, minsize - len(output))

	return tuple((*output, *filler))


def count(start: AnyNum = 0, step: AnyNum = 1) -> Iterator[AnyNum]:
	"""
	Make an iterator which returns evenly spaced values starting with number ``start``.

	Often used as an argument to :func:`map` to generate consecutive data points.
	Can also be used with :func:`zip` to add sequence numbers.

	.. versionadded:: 2.7.0

	:param start:
	:param step: The step between values.

	:rtype:

	.. seealso::

		:func:`itertools.count`.

		The difference is that this returns more exact floats,
		whereas the values from :func:`itertools.count` drift.

		.. only:: html

			A demonstration of the drift can be seen in this file: :download:`count_demo.py`.

	.. latex:clearpage::
	"""

	if not isinstance(start, (int, float, complex)):
		raise TypeError("a number is required")
	if not isinstance(step, (int, float, complex)):
		raise TypeError("a number is required")

	# count(10) --> 10 11 12 13 14 ...
	# count(2.5, 0.5) -> 2.5 3.0 3.5 ...

	pos: int = 0

	def get_next():
		if pos:
			return start + (step * pos)
		else:
			return start

	@final
	class count(Iterator[AnyNum]):

		def __next__(self):
			nonlocal pos

			val = get_next()
			pos += 1

			return val

		def __iter__(self):
			return self

		if isinstance(step, int) and step == 1:

			def __repr__(self) -> str:
				return f"{self.__class__.__name__}({get_next()})"
		else:

			def __repr__(self) -> str:
				return f"{self.__class__.__name__}{get_next(), step}"

		def __init_subclass__(cls, **kwargs):
			raise TypeError("type 'domdf_python_tools.iterative.count' is not an acceptable base type")

	count.__qualname__ = count.__name__ = "count"

	return count()  # type: ignore



#!/usr/bin/env python
#
#  getters.py
"""
Variants of :func:`operator.attrgetter`, :func:`operator.itemgetter` and :func:`operator.methodcaller`
which operate on values within sequences.

.. versionadded:: 3.2.0
"""  # noqa: D400
#
#  Copyright © 2021 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#
#  Adapted from https://github.com/python/cpython/blob/master/Lib/operator.py
#  Licensed under the Python Software Foundation License Version 2.
#  Copyright © 2001-2020 Python Software Foundation. All rights reserved.
#  Copyright © 2000 BeOpen.com. All rights reserved.
#  Copyright © 1995-2000 Corporation for National Research Initiatives. All rights reserved.
#  Copyright © 1991-1995 Stichting Mathematisch Centrum. All rights reserved.
#

# stdlib
from functools import partial
from typing import TYPE_CHECKING, Any, Dict, Tuple

__all__ = ["attrgetter", "itemgetter", "methodcaller"]


class attrgetter:
	"""
	Returns a callable object that fetches ``attr`` from the item at index ``idx`` in its operand.

	The attribute name can contain dots. For example:

	* After ``f = attrgetter(0, 'name')``, the call call ``f(b)`` returns ``b[0].name``.
	* After ``f = attrgetter(3, 'name.first')``, the call ``f(b)`` returns ``b[3].name.first``.

	.. code-block:: python

		>>> from pathlib import Path
		>>> attrgetter(0, 'name')([Path("dir/code.py")])
		'code.py'
		>>> attrgetter(2, 'parent.name')([Path("dir/coincidence.py"), Path("dir/wheel.py"), Path("dir/operator.py")])
		'dir'

	.. seealso:: :func:`operator.attrgetter` and :func:`operator.itemgetter`

	:param idx: The index of the item to obtain the attribute from.
	:param attr: The name of the attribute.
	"""

	__slots__ = ("_attrs", "_call")

	def __init__(self, idx: int, attr: str):
		if not isinstance(idx, int):
			raise TypeError("'idx' must be an integer")

		if not isinstance(attr, str):
			raise TypeError("attribute name must be a string")

		self._attrs: Dict[str, Any] = {"idx": idx, "attr": attr}

	def __call__(self, obj: Any) -> Any:  # noqa: D102
		names = self._attrs["attr"].split('.')

		obj = obj[self._attrs["idx"]]

		for name in names:
			obj = getattr(obj, name)

		return obj

	def __repr__(self) -> str:
		data = {**self._attrs, "module": self.__class__.__module__, "qualname": self.__class__.__qualname__}
		return "{module}.{qualname}(idx={idx}, attr={attr!r})".format_map(data)

	def __reduce__(self):
		return self.__class__, tuple(self._attrs.values())


class itemgetter:
	"""
	Returns a callable object that fetches ``item`` from the item at index ``idx`` in its operand,
	using the ``__getitem__()`` method.

	For example:

	* After ``f = itemgetter(0, 2)``, the call call ``f(r)`` returns ``r[0][2]``.
	* After ``g = itemgetter(3, 5)``, the call ``g(r)`` returns ``r[3][5]``.

	The items can be any type accepted by the item's ``__getitem__()`` method.
	Dictionaries accept any hashable value. Lists, tuples, and strings accept an index or a slice:

	.. code-block:: python

		>>> itemgetter(0, 1)(['ABCDEFG'])
		'B'
		>>> itemgetter(1, 2)(['ABC', 'DEF'])
		'F'
		>>> itemgetter(0, slice(2, None))(['ABCDEFG'])
		'CDEFG'
		>>> army = [dict(rank='captain', name='Blackadder'), dict(rank='Private', name='Baldrick')]
		>>> itemgetter(0, 'rank')(army)
		'captain'

	.. seealso:: :func:`operator.itemgetter`

	:param idx: The index of the item to call ``__getitem__()`` on.
	:param item: The value to pass to ``__getitem__()``.
	"""  # noqa: D400

	__slots__ = ("_items", "_call")

	def __init__(self, idx: int, item: Any):
		if not isinstance(idx, int):
			raise TypeError("'idx' must be an integer")

		self._items: Dict[str, Any] = {"idx": idx, "item": item}

	def __call__(self, obj: Any) -> Any:  # noqa: D102
		return obj[self._items["idx"]][self._items["item"]]

	def __repr__(self) -> str:
		data = {**self._items, "module": self.__class__.__module__, "qualname": self.__class__.__qualname__}
		return "{module}.{qualname}(idx={idx}, item={item!r})".format_map(data)

	def __reduce__(self):
		return self.__class__, tuple(self._items.values())


class methodcaller:
	r"""
	Returns a callable object that calls the method name on the item at index ``idx`` in its operand.

	If additional arguments and/or keyword arguments are given, they will be passed to the method as well.
	For example:

	* After ``f = methodcaller(0, 'name')``, the call ``f(b)`` returns ``b[0].name()``.
	* After ``f = methodcaller(1, 'name', 'foo', bar=1)``, the call ``f(b)`` returns ``b[1].name('foo', bar=1)``.

	.. code-block:: python

		>>> from datetime import date
		>>> methodcaller(0, 'upper')(["hello", "world"])
		'HELLO'
		>>> methodcaller(1, 'center', 9, "=")(["hello", "world"])
		'==world=='
		>>> methodcaller(0, 'replace', year=2019)([date(2021, 7, 6)])
		datetime.date(2019, 7, 6)

	.. seealso:: :func:`operator.methodcaller` and :func:`operator.itemgetter`

	:param \_idx: The index of the item to call the method on.
	:param \_attr: The name of the method to call.
	:param \*args: Positional arguments to pass to the method.
	:param \*\*kwargs: Keyword arguments to pass to the method.
	"""

	__slots__ = ("_idx", "_name", "_args", "_kwargs")

	_idx: int
	_name: str
	_args: Tuple[Any, ...]
	_kwargs: Dict[str, Any]

	if TYPE_CHECKING:

		def __init__(__self, __idx: int, __name: str, *args, **kwargs):
			if not isinstance(__idx, int):
				raise TypeError("'idx' must be an integer")

			if not isinstance(__name, str):
				raise TypeError("method name must be a string")

			__self._idx = __idx
			__self._name = __name
			__self._args = args
			__self._kwargs = kwargs

	else:

		def __init__(_self, _idx: int, _name: str, *args, **kwargs):
			if not isinstance(_idx, int):
				raise TypeError("'_idx' must be an integer")

			if not isinstance(_name, str):
				raise TypeError("method name must be a string")

			_self._idx = _idx
			_self._name = _name
			_self._args = args
			_self._kwargs = kwargs

	def __call__(self, obj: Any) -> Any:  # noqa: D102
		return getattr(obj[self._idx], self._name)(*self._args, **self._kwargs)

	def __repr__(self) -> str:
		args = [repr(self._idx), repr(self._name)]
		args.extend(map(repr, self._args))
		args.extend(f'{k}={v!r}' for k, v in self._kwargs.items())
		return f'{self.__class__.__module__}.{self.__class__.__name__}({", ".join(args)})'

	def __reduce__(self):
		if not self._kwargs:
			return self.__class__, (self._idx, self._name) + self._args
		else:
			return partial(self.__class__, self._idx, self._name, **self._kwargs), self._args



#  !/usr/bin/env python
#
#  doctools.py
"""
Utilities for documenting functions, classes and methods.

.. autosummary-widths:: 5/16

.. automodulesumm:: domdf_python_tools.doctools
	:autosummary-sections: Data

.. autosummary-widths:: 17/32

.. automodulesumm:: domdf_python_tools.doctools
	:autosummary-sections: Functions
"""
#
#  Copyright © 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#  Based on https://softwareengineering.stackexchange.com/a/386758
#  Copyright © amon (https://softwareengineering.stackexchange.com/users/60357/amon)
#  Licensed under CC BY-SA 4.0
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
import builtins
from contextlib import suppress
from inspect import cleandoc
from types import MethodType
from typing import Any, Callable, Dict, Optional, Sequence, Type, TypeVar, Union

# this package
from domdf_python_tools.compat import PYPY, PYPY37
from domdf_python_tools.typing import MethodDescriptorType, MethodWrapperType, WrapperDescriptorType

__all__ = [
		"_F",
		"_T",
		"deindent_string",
		"document_object_from_another",
		"append_doctring_from_another",
		"make_sphinx_links",
		"is_documented_by",
		"append_docstring_from",
		"sphinxify_docstring",
		"prettify_docstrings",
		]

_F = TypeVar("_F", bound=Callable[..., Any])
_T = TypeVar("_T", bound=Type)


def deindent_string(string: Optional[str]) -> str:
	"""
	Removes all indentation from the given string.

	:param string: The string to deindent

	:return: The string without indentation
	"""

	if not string:
		# Short circuit if empty string or None
		return ''

	split_string = string.split('\n')
	deindented_string = [line.lstrip("\t ") for line in split_string]
	return '\n'.join(deindented_string)


# Functions that do the work
def document_object_from_another(target: Union[Type, Callable], original: Union[Type, Callable]):
	"""
	Sets the docstring of the ``target`` function to that of the ``original`` function.

	This may be useful for subclasses or wrappers that use the same arguments.

	:param target: The object to set the docstring for
	:param original: The object to copy the docstring from
	"""

	target.__doc__ = original.__doc__


def append_doctring_from_another(target: Union[Type, Callable], original: Union[Type, Callable]):
	"""
	Sets the docstring of the ``target`` function to that of the ``original`` function.

	This may be useful for subclasses or wrappers that use the same arguments.

	Any indentation in either docstring is removed to
	ensure consistent indentation between the two docstrings.
	Bear this in mind if additional indentation is used in the docstring.

	:param target: The object to append the docstring to
	:param original: The object to copy the docstring from
	"""

	# this package
	from domdf_python_tools.stringlist import StringList

	target_doc = target.__doc__
	original_doc = original.__doc__

	if isinstance(original_doc, str) and isinstance(target_doc, str):
		docstring = StringList(cleandoc(target_doc))
		docstring.blankline(ensure_single=True)
		docstring.append(cleandoc(original_doc))
		docstring.blankline(ensure_single=True)
		target.__doc__ = str(docstring)

	elif not isinstance(target_doc, str) and isinstance(original_doc, str):
		docstring = StringList(cleandoc(original_doc))
		docstring.blankline(ensure_single=True)
		target.__doc__ = str(docstring)


def make_sphinx_links(input_string: str, builtins_list: Optional[Sequence[str]] = None) -> str:
	r"""
	Make proper sphinx links out of double-backticked strings in docstring.

	i.e. :inline-code:`\`\`str\`\`` becomes  :inline-code:`:class:\`str\``

	Make sure to include the following in your ``conf.py`` file for Sphinx:

	.. code-block:: python

		intersphinx_mapping = {"python": ("https://docs.python.org/3/", None)}

	:param input_string: The string to process.
	:param builtins_list: A list of builtins to make links for.
	:default builtins_list: dir(:py:obj:`builtins`)

	:return: Processed string with links.
	"""

	if builtins_list is None:
		builtins_list = dir(builtins)

	working_string = f"{input_string}"

	for builtin in builtins_list:
		if builtin.startswith("__"):
			continue

		if builtin in {"None", "False", "None"}:
			working_string = working_string.replace(f"``{builtin}``", f":py:obj:`{builtin}`")
		else:
			working_string = working_string.replace(f"``{builtin}``", f":class:`{builtin}`")

	return working_string


# Decorators that call the above functions
def is_documented_by(original: Callable) -> Callable[[_F], _F]:
	"""
	Decorator to set the docstring of the ``target`` function to that of the ``original`` function.

	This may be useful for subclasses or wrappers that use the same arguments.

	:param original:
	"""

	def wrapper(target: _F) -> _F:
		document_object_from_another(target, original)
		return target

	return wrapper


def append_docstring_from(original: Callable) -> Callable[[_F], _F]:
	"""
	Decorator to appends the docstring from the ``original`` function to the ``target`` function.

	This may be useful for subclasses or wrappers that use the same arguments.

	Any indentation in either docstring is removed to
	ensure consistent indentation between the two docstrings.
	Bear this in mind if additional indentation is used in the docstring.

	:param original:
	"""

	def wrapper(target: _F) -> _F:
		append_doctring_from_another(target, original)
		return target

	return wrapper


def sphinxify_docstring() -> Callable[[_F], _F]:
	r"""
	Decorator to make proper sphinx links out of double-backticked strings in the docstring.

	i.e. :inline-code:`\`\`str\`\`` becomes  :inline-code:`:class:\`str\``

	Make sure to include the following in your ``conf.py`` file for Sphinx:

	.. code-block:: python

		intersphinx_mapping = {
			"python": ("https://docs.python.org/3/", None),
		}
	"""

	def wrapper(target: _F) -> _F:
		target_doc = target.__doc__

		if target_doc:
			target.__doc__ = make_sphinx_links(target_doc)

		return target

	return wrapper


# Check against object
base_new_docstrings = {
		"__delattr__": "Implement :func:`delattr(self, name) <delattr>`.",
		"__dir__": "Default :func:`dir` implementation.",
		"__eq__": "Return ``self == other``.",  # __format__
		"__getattribute__": "Return :func:`getattr(self, name) <getattr>`.",
		"__ge__": "Return ``self >= other``.",
		"__gt__": "Return ``self > other``.",
		"__hash__": "Return :func:`hash(self) <hash>`.",
		# __init_subclass__
		# __init__  # not usually shown in sphinx
		"__lt__": "Return ``self < other``.",
		"__le__": "Return ``self <= other``.",  # __new__
		"__ne__": "Return ``self != other``.",
		# __reduce_ex__
		# __reduce__
		# __repr__ is defined within the function
		"__setattr__": "Implement :func:`setattr(self, name) <setattr>`.",
		"__sizeof__": "Returns the size of the object in memory, in bytes.",
		"__str__": "Return :class:`str(self) <str>`.",  # __subclasshook__
		}

# Check against dict
container_docstrings = {
		"__contains__": "Return ``key in self``.",
		"__getitem__": "Return ``self[key]``.",
		"__setitem__": "Set ``self[key]`` to ``value``.",
		"__delitem__": "Delete ``self[key]``.",
		}

# Check against int
operator_docstrings = {
		"__and__": "Return ``self & value``.",
		"__add__": "Return ``self + value``.",
		"__abs__": "Return :func:`abs(self) <abs>`.",
		"__divmod__": "Return :func:`divmod(self, value) <divmod>`.",
		"__floordiv__": "Return ``self // value``.",
		"__invert__": "Return ``~ self``.",
		"__lshift__": "Return ``self << value``.",
		"__mod__": "Return ``self % value``.",
		"__mul__": "Return ``self * value``.",
		"__neg__": "Return ``- self``.",
		"__or__": "Return ``self | value``.",
		"__pos__": "Return ``+ self``.",
		"__pow__": "Return :func:`pow(self, value, mod) <pow>`.",
		"__radd__": "Return ``value + self``.",
		"__rand__": "Return ``value & self``.",
		"__rdivmod__": "Return :func:`divmod(value, self) <divmod>`.",
		"__rfloordiv__": "Return ``value // self``.",
		"__rlshift__": "Return ``value << self``.",
		"__rmod__": "Return ``value % self``.",
		"__rmul__": "Return ``value * self``.",
		"__ror__": "Return ``value | self``.",
		"__rpow__": "Return :func:`pow(value, self, mod) <pow>`.",
		"__rrshift__": "Return ``self >> value``.",
		"__rshift__": "Return ``self >> value``.",
		"__rsub__": "Return ``value - self``.",
		"__rtruediv__": "Return ``value / self``.",
		"__rxor__": "Return ``value ^ self``.",
		"__sub__": "Return ``value - self``.",
		"__truediv__": "Return ``self / value``.",
		"__xor__": "Return ``self ^ value``.",
		}

# Check against int
base_int_docstrings = {
		# "__bool__": "Return ``self != 0``.",  # TODO
		# __ceil__
		"__float__": "Return :class:`float(self) <float>`.",  # __floor__
		"__int__": "Return :class:`int(self) <int>`.",  # __round__
		}

new_return_types = {
		"__eq__": bool,
		"__ge__": bool,
		"__gt__": bool,
		"__lt__": bool,
		"__le__": bool,
		"__ne__": bool,
		"__repr__": str,
		"__str__": str,
		"__int__": int,
		"__float__": float,
		"__bool__": bool,
		}


def _do_prettify(obj: Type, base: Type, new_docstrings: Dict[str, str]):
	"""
	Perform the actual prettifying for :func`~.prettify_docstrings`.

	.. versionadded:: 0.8.0 (private)

	:param obj:
	:param base:
	:param new_docstrings:
	"""

	for attr_name in new_docstrings:

		if not hasattr(obj, attr_name):
			continue

		attribute = getattr(obj, attr_name)

		if not PYPY and isinstance(
				attribute,
				(WrapperDescriptorType, MethodDescriptorType, MethodWrapperType, MethodType),
				):
			continue  # pragma: no cover (!PyPy)
		elif PYPY and isinstance(attribute, MethodType):
			continue  # pragma: no cover
		elif PYPY37:  # pragma: no cover (not (PyPy and py37))
			if attribute is getattr(object, attr_name, None):
				continue
			elif attribute is getattr(float, attr_name, None):
				continue
			elif attribute is getattr(str, attr_name, None):
				continue

		if attribute is None:
			continue

		base_docstring: Optional[str] = None
		if hasattr(base, attr_name):
			base_docstring = getattr(base, attr_name).__doc__

		doc: Optional[str] = attribute.__doc__
		if doc in {None, base_docstring}:
			with suppress(AttributeError, TypeError):
				attribute.__doc__ = new_docstrings[attr_name]


def prettify_docstrings(obj: _T) -> _T:
	"""
	Decorator to prettify the default :class:`object` docstrings for use in Sphinx documentation.

	.. versionadded:: 0.8.0

	:param obj: The object to prettify the method docstrings for.
	"""

	repr_docstring = f"Return a string representation of the :class:`~{obj.__module__}.{obj.__name__}`."
	new_docstrings = {**base_new_docstrings, "__repr__": repr_docstring}

	_do_prettify(obj, object, new_docstrings)
	_do_prettify(obj, dict, container_docstrings)
	_do_prettify(obj, int, operator_docstrings)
	_do_prettify(obj, int, base_int_docstrings)

	for attribute in new_return_types:
		if hasattr(obj, attribute):
			annotations: Dict = getattr(getattr(obj, attribute), "__annotations__", {})

			if "return" not in annotations or annotations["return"] is Any:
				annotations["return"] = new_return_types[attribute]

			with suppress(AttributeError, TypeError):
				getattr(obj, attribute).__annotations__ = annotations

	if issubclass(obj, tuple) and obj.__repr__.__doc__ == "Return a nicely formatted representation string":
		obj.__repr__.__doc__ = repr_docstring

	return obj



#  !/usr/bin/env python
#
#  classes.py
"""
Classes representing pagesizes.
"""
#
#  Copyright © 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Based on reportlab.lib.pagesizes and reportlab.lib.units
#    www.reportlab.co.uk
#    Copyright ReportLab Europe Ltd. 2000-2017
#    Copyright (c) 2000-2018, ReportLab Inc.
#    All rights reserved.
#    Licensed under the BSD License
#
#  Includes data from en.wikipedia.org.
#  Licensed under the Creative Commons Attribution-ShareAlike License
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
from collections import namedtuple
from typing import List, Tuple

# this package
from domdf_python_tools.doctools import prettify_docstrings
from domdf_python_tools.typing import AnyNumber

# this package
from .units import Unit, _rounders, cm, inch, mm, pica, pt, um
from .utils import convert_from

__all__ = [
		"BaseSize",
		"Size_mm",
		"Size_inch",
		"Size_cm",
		"Size_um",
		"Size_pica",
		"PageSize",
		]


@prettify_docstrings
class BaseSize(namedtuple("__BaseSize", "width, height")):
	"""
	Base class namedtuple representing a page size, in point.
	"""

	__slots__: List[str] = []
	_unit: Unit = pt

	#: The page width.
	width: Unit

	#: The page height.
	height: Unit

	def __new__(cls, width: AnyNumber, height: AnyNumber):
		"""
		Create a new :class:`~.BaseSize` object.

		:param width: The page width.
		:param height: The page height.
		"""

		return super().__new__(
				cls,
				cls._unit(width),
				cls._unit(height),
				)

	def __str__(self) -> str:
		return f"{self.__class__.__name__}(width={_rounders(self.width, '0')}, height={_rounders(self.height, '0')})"

	@classmethod
	def from_pt(cls, size: Tuple[float, float]):
		"""
		Create a :class:`~domdf_python_tools.pagesizes.classes.BaseSize` object from a
		page size in point.

		:param size: The size, in point, to convert from.

		:rtype: A subclass of :class:`~domdf_python_tools.pagesizes.classes.BaseSize`
		"""  # noqa: D400

		assert isinstance(size, PageSize)

		return cls(cls._unit.from_pt(size[0]), cls._unit.from_pt(size[1]))

	@classmethod
	def from_size(cls, size: Tuple[AnyNumber, AnyNumber]) -> "BaseSize":
		"""
		Create a :class:`~domdf_python_tools.pagesizes.classes.BaseSize` object from a tuple.
		"""

		return cls(*size)

	def is_landscape(self) -> bool:
		"""
		Returns whether the page is in the landscape orientation.
		"""

		return self.width >= self.height

	def is_portrait(self) -> bool:
		"""
		Returns whether the page is in the portrait orientation.
		"""

		return self.width < self.height

	def is_square(self) -> bool:
		"""
		Returns whether the given pagesize is square.
		"""

		return self.width == self.height

	def landscape(self) -> "BaseSize":
		"""
		Returns the pagesize in landscape orientation.
		"""

		if self.is_portrait():
			return self.__class__(self.height, self.width)
		else:
			return self

	def portrait(self) -> "BaseSize":
		"""
		Returns the pagesize in portrait orientation.
		"""

		if self.is_landscape():
			return self.__class__(self.height, self.width)
		else:
			return self

	def to_pt(self) -> "PageSize":
		"""
		Returns the page size in point.
		"""

		return PageSize(self.width.as_pt(), self.height.as_pt())


# TODO: conversion to Point for the __eq__ function in the below


class Size_mm(BaseSize):
	"""
	Represents a pagesize in millimeters.
	"""

	_unit = mm


class Size_inch(BaseSize):
	"""
	Represents a pagesize in inches.
	"""

	_unit = inch


class Size_cm(BaseSize):
	"""
	Represents a pagesize in centimeters.
	"""

	_unit = cm


class Size_um(BaseSize):
	"""
	Represents a pagesize in micrometers.
	"""

	_unit = um


class Size_pica(BaseSize):
	"""
	Represents a pagesize in pica.
	"""

	_unit = pica


class PageSize(BaseSize):
	"""
	Represents a pagesize in point.

	:param width: The page width
	:param height: The page height

	The pagesize can be converted to other units using the properties below.
	"""

	__slots__: List[str] = []

	def __new__(
			cls,
			width: AnyNumber,
			height: AnyNumber,
			unit: AnyNumber = pt,  # pylint: disable=used-before-assignment
			):
		"""
		Create a new :class:`~domdf_python_tools.pagesizes.classes.PageSize` object.

		:param width: The page width.
		:param height: The page height.
		:param unit:
		"""

		width, height = convert_from((width, height), unit)
		return super().__new__(cls, width, height)

	@property
	def pt(self) -> "PageSize":
		"""
		Returns the pagesize in pt.
		"""

		return self

	@property
	def inch(self) -> Size_inch:
		"""
		Returns the pagesize in inches.
		"""

		return Size_inch.from_pt(self)

	@property
	def cm(self) -> Size_cm:
		"""
		Returns the pagesize in centimeters.
		"""

		return Size_cm.from_pt(self)

	@property
	def mm(self) -> Size_mm:
		"""
		Returns the pagesize in millimeters.
		"""

		return Size_mm.from_pt(self)

	@property
	def um(self) -> Size_um:
		"""
		Returns the pagesize in micrometers.
		"""

		return Size_um.from_pt(self)

	µm = um

	@property
	def pc(self) -> Size_pica:
		"""
		Returns the pagesize in pica.
		"""

		return Size_pica.from_pt(self)

	pica = pc



#!/usr/bin/env python
#
#  typing.py
"""
Various type annotation aids.
"""
#
#  Copyright © 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
import os
import pathlib
from decimal import Decimal
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Tuple, Type, TypeVar, Union

# 3rd party
from typing_extensions import Protocol, runtime_checkable

# this package
import domdf_python_tools

if TYPE_CHECKING or domdf_python_tools.__docs:  # pragma: no cover
	# stdlib
	from json import JSONDecoder, JSONEncoder

	# 3rd party
	from pandas import DataFrame, Series

	Series.__module__ = "pandas"
	DataFrame.__module__ = "pandas"

	JSONDecoder.__module__ = "json"
	JSONEncoder.__module__ = "json"

#: .. versionadded:: 1.0.0
FrameOrSeries = TypeVar("FrameOrSeries", "Series", "DataFrame")

__all__ = [
		"PathLike",
		"PathType",
		"AnyNumber",
		"WrapperDescriptorType",
		"MethodWrapperType",
		"MethodDescriptorType",
		"ClassMethodDescriptorType",
		"JsonLibrary",
		"HasHead",
		"String",
		"FrameOrSeries",
		"SupportsIndex",
		"SupportsLessThan",
		"SupportsLessEqual",
		"SupportsGreaterThan",
		"SupportsGreaterEqual",
		"check_membership",
		]

PathLike = Union[str, pathlib.Path, os.PathLike]
"""
Type hint for objects that represent filesystem paths.

.. seealso:: :py:obj:`domdf_python_tools.typing.PathType`
"""

PathType = TypeVar("PathType", str, pathlib.Path, os.PathLike)
"""
Type variable for objects that represent filesystem paths.

.. versionadded:: 2.2.0

.. seealso:: :py:obj:`domdf_python_tools.typing.PathLike`
"""

AnyNumber = Union[float, int, Decimal]
"""
Type hint for numbers.

.. versionchanged:: 0.4.6

	Moved from :mod:`domdf_python_tools.pagesizes`
"""


def check_membership(obj: Any, type_: Union[Type, object]) -> bool:
	r"""
	Check if the type of ``obj`` is one of the types in a :py:data:`typing.Union`, :class:`typing.Sequence` etc.

	:param obj: The object to check the type of
	:param type\_: A :class:`~typing.Type` that has members,
		such as a :class:`typing.List`, :py:data:`typing.Union` or :py:class:`typing.Sequence`.
	"""

	return isinstance(obj, type_.__args__)  # type: ignore


class JsonLibrary(Protocol):
	"""
	:class:`typing.Protocol` for libraries that implement the same API as :mod:`json`.

	Useful for annotating functions which take a JSON serialisation-deserialisation library as an argument.
	"""

	@staticmethod
	def dumps(
			obj: Any,
			*,
			skipkeys: bool = ...,
			ensure_ascii: bool = ...,
			check_circular: bool = ...,
			allow_nan: bool = ...,
			cls: Optional[Type["JSONEncoder"]] = ...,
			indent: Union[None, int, str] = ...,
			separators: Optional[Tuple[str, str]] = ...,
			default: Optional[Callable[[Any], Any]] = ...,
			sort_keys: bool = ...,
			**kwds: Any,
			) -> str:
		"""
		Serialize ``obj`` to a JSON formatted :class:`str`.

		:param obj:
		:param skipkeys:
		:param ensure_ascii:
		:param check_circular:
		:param allow_nan:
		:param cls:
		:param indent:
		:param separators:
		:param default:
		:param sort_keys:
		:param kwds:
		"""

	@staticmethod
	def loads(
			s: Union[str, bytes],
			*,
			cls: Optional[Type["JSONDecoder"]] = ...,
			object_hook: Optional[Callable[[Dict[Any, Any]], Any]] = ...,
			parse_float: Optional[Callable[[str], Any]] = ...,
			parse_int: Optional[Callable[[str], Any]] = ...,
			parse_constant: Optional[Callable[[str], Any]] = ...,
			object_pairs_hook: Optional[Callable[[List[Tuple[Any, Any]]], Any]] = ...,
			**kwds: Any,
			) -> Any:
		"""
		Deserialize ``s`` to a Python object.

		:param s:
		:param cls:
		:param object_hook:
		:param parse_float:
		:param parse_int:
		:param parse_constant:
		:param object_pairs_hook:
		:param kwds:

		:rtype:

		.. latex:clearpage::
		"""


#  Backported from https://github.com/python/cpython/blob/master/Lib/types.py
#  Licensed under the Python Software Foundation License Version 2.
#  Copyright © 2001-2020 Python Software Foundation. All rights reserved.
#  Copyright © 2000 BeOpen.com. All rights reserved.
#  Copyright © 1995-2000 Corporation for National Research Initiatives. All rights reserved.
#  Copyright © 1991-1995 Stichting Mathematisch Centrum. All rights reserved.

WrapperDescriptorType = type(object.__init__)
MethodWrapperType = type(object().__str__)
MethodDescriptorType = type(str.join)
ClassMethodDescriptorType = type(dict.__dict__["fromkeys"])


@runtime_checkable
class String(Protocol):
	"""
	:class:`~typing.Protocol` for classes that implement ``__str__``.

	.. versionchanged:: 0.8.0

		Moved from :mod:`domdf_python_tools.stringlist`.
	"""

	def __str__(self) -> str: ...


@runtime_checkable
class HasHead(Protocol):
	"""
	:class:`typing.Protocol` for classes that have a ``head`` method.

	This includes :class:`pandas.DataFrame` and :class:`pandas.Series`.

	.. versionadded:: 0.8.0
	"""

	def head(self: "HasHead", n: int = 5) -> "HasHead":
		"""
		Return the first n rows.

		:param n: Number of rows to select.

		:return: The first n rows of the caller object.
		"""

	def to_string(self, *args, **kwargs) -> Optional[str]:
		"""
		Render the object to a console-friendly tabular output.
		"""


# class SupportsLessThan(Protocol):
#
# 	def __lt__(self, other: Any) -> bool:
# 		...  # pragma: no cover


class SupportsIndex(Protocol):
	"""
	:class:`typing.Protocol` for classes that support ``__index__``.

	.. versionadded:: 2.0.0
	"""

	def __index__(self) -> int: ...


class SupportsLessThan(Protocol):
	"""
	:class:`typing.Protocol` for classes that support ``__lt__``.

	.. versionadded:: 3.0.0
	"""

	def __lt__(self, __other: Any) -> bool:
		"""
		Return ``self < value``.
		"""


class SupportsLessEqual(Protocol):
	"""
	:class:`typing.Protocol` for classes that support ``__le__``.

	.. versionadded:: 3.0.0
	"""

	def __le__(self, __other: Any) -> bool:
		"""
		Return ``self <= value``.
		"""


class SupportsGreaterThan(Protocol):
	"""
	:class:`typing.Protocol` for classes that support ``__gt__``.

	.. versionadded:: 3.0.0
	"""

	def __gt__(self, __other: Any) -> bool:
		"""
		Return ``self > value``.
		"""


class SupportsGreaterEqual(Protocol):
	"""
	:class:`typing.Protocol` for classes that support ``__ge__``.

	.. versionadded:: 3.0.0
	"""

	def __ge__(self, __other: Any) -> bool:
		"""
		Return ``self >= value``.
		"""



#!/usr/bin/env python
#
#  units.py
"""
Provides a variety of units for use with pagesizes.
"""
#
#  Copyright © 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Based on reportlab.lib.pagesizes and reportlab.lib.units
#    www.reportlab.co.uk
#    Copyright ReportLab Europe Ltd. 2000-2017
#    Copyright (c) 2000-2018, ReportLab Inc.
#    All rights reserved.
#    Licensed under the BSD License
#
#  Includes data from en.wikipedia.org.
#  Licensed under the Creative Commons Attribution-ShareAlike License
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
import math
from decimal import ROUND_HALF_UP, Decimal
from typing import SupportsFloat, Union

# this package
from domdf_python_tools.doctools import prettify_docstrings

__all__ = [
		"pt",
		"inch",
		"cm",
		"mm",
		"um",
		"pc",
		"pica",
		"Unit",
		"Unitpt",
		"UnitInch",
		"Unitcm",
		"Unitmm",
		"Unitum",
		"Unitpc",
		]


def _rounders(val_to_round: Union[str, int, float, Decimal], round_format: str) -> Decimal:
	return Decimal(Decimal(val_to_round).quantize(Decimal(str(round_format)), rounding=ROUND_HALF_UP))


@prettify_docstrings
class Unit(float):
	r"""
	Represents a unit, such as a point.

	Behaves much like a float (which it inherits from).


	:bold-title:`Addition`

	Units can be added to each other:

	.. code-block:: python

		>>> (3*mm) + (7*mm)
		<Unit '10.000 mm': 28.346pt>

	When adding different :class:`~domdf_python_tools.pagesizes.units.Unit` objects,
	the result has the type of the former unit:

	.. code-block:: python

		>>> (2.54*cm) + inch
		<Unit '5.080 cm': 144.000pt>
		>>> inch + (2.54*cm)
		<Unit '2.000 inch': 144.000pt>

	:class:`~domdf_python_tools.pagesizes.units.Unit` objects can also be added to :class:`float` and :class:`int` objects:

	.. code-block:: python

		>>> (3*cm) + 7
		<Unit '10.000 cm': 283.465pt>
		>>> 7 + (3*cm)
		<Unit '10.000 cm': 283.465pt>


	:bold-title:`Subtraction`

	Subtraction works the same as addition:

	.. code-block:: python

		>>> (17*mm) - (7*mm)
		<Unit '10.000 mm': 28.346pt>
		>>> (2.54*cm) - inch
		<Unit '0.000 cm': 0.000pt>
		>>> inch - (2.54*cm)
		<Unit '0.000 inch': 0.000pt>
		>>> (17*cm) - 7
		<Unit '10.000 cm': 283.465pt>
		>>> 17 - (7*cm)
		<Unit '10.000 cm': 283.465pt>


	:bold-title:`Multiplication`

	:class:`~domdf_python_tools.pagesizes.units.Unit` objects can only be multipled by
	:class:`float` and :class:`int` objects:

	.. code-block:: python

		>>> (3*mm) * 3
		<Unit '9.000 mm': 25.512pt>
		>>> 3 * (3*mm)
		<Unit '9.000 mm': 25.512pt>
		>>> 3.5 * (3*mm)
		<Unit '10.500 mm': 29.764pt>

	Multiplication works either way round.

	Multiplying by another :class:`~domdf_python_tools.pagesizes.units.Unit`
	results in a :exc:`NotImplementedError`:

	.. code-block:: python

		>>> inch * (7*cm)
		Traceback (most recent call last):
		NotImplementedError: Multiplying a unit by another unit is not allowed.

	:bold-title:`Division`

	:class:`~domdf_python_tools.pagesizes.units.Unit`\s can only be divided by
	:class:`float` and :class:`int` objects:

	.. code-block:: python

		>>> (3*mm) / 3
		<Unit '1.000 mm': 2.835pt>
		>>> (10*mm) / 2.5
		<Unit '4.000 mm': 11.339pt>

	Dividing by another unit results in a :exc:`NotImplementedError`:

	.. code-block:: python

		>>> inch / (7*cm)
		Traceback (most recent call last):
		NotImplementedError: Dividing a unit by another unit is not allowed.

	Likewise, trying to divide a:class:`float` and :class:`int` object by a unit
	results in a :exc:`NotImplementedError`:

	.. code-block:: python

		>>> 3 / (3*mm)
		Traceback (most recent call last):
		NotImplementedError: Dividing by a unit is not allowed.


	:bold-title:`Powers`

	Powers (using ``**``) are not officially supported.


	:bold-title:`Modulo Division`

	Modulo division of a :class:`~domdf_python_tools.pagesizes.units.Unit` by a
	:class:`float` or :class:`int` object is allowed:

	.. code-block:: python

		>>> (3*mm) % 2.5
		<Unit '0.500 mm': 1.417pt>

	Dividing by a unit, or modulo division of two units, is not officially supported.

	.. latex:clearpage::
	"""

	name: str = "pt"
	_in_pt: float = 1

	def __repr__(self):
		value = _rounders(float(self), "0.000")
		as_pt = _rounders(self.as_pt(), "0.000")
		return f"<Unit '{value} {self.name}': {as_pt}pt>"

	def __str__(self):
		value = _rounders(float(self), "0.000")
		as_pt = _rounders(self.as_pt(), "0.000")
		return f"<Unit '{value}\u205F{self.name}': {as_pt}pt>"

	def __mul__(self, other: Union[float, "Unit"]) -> "Unit":
		if isinstance(other, Unit):
			raise NotImplementedError("Multiplying a unit by another unit is not allowed.")

		return self.__class__(super().__mul__(other))

	__rmul__ = __mul__

	def __truediv__(self, other: Union[float, "Unit"]) -> "Unit":
		if isinstance(other, Unit):
			raise NotImplementedError("Dividing a unit by another unit is not allowed.")

		return self.__class__(super().__truediv__(other))

	def __floordiv__(self, other: Union[float, "Unit"]) -> "Unit":
		if isinstance(other, Unit):
			raise NotImplementedError("Dividing a unit by another unit is not allowed.")

		return self.__class__(super().__floordiv__(other))

	def __eq__(self, other: object) -> bool:
		if isinstance(other, Unit):
			if self._in_pt != 1:
				self_value = self.as_pt()
			else:
				self_value = self

			if other._in_pt != 1:
				other_value = other.as_pt()
			else:
				other_value = other

			return math.isclose(float(self_value), float(other_value), abs_tol=1e-8)
		else:
			return super().__eq__(other)

	def __mod__(self, other: Union[float, "Unit"]) -> "Unit":
		if isinstance(other, Unit):
			raise NotImplementedError("Modulo division of a unit by another unit is not allowed.")

		return self.__class__(super().__mod__(other))

	def __pow__(self, power, modulo=None):
		raise NotImplementedError("Powers are not supported for units.")

	def __rtruediv__(self, other):
		raise NotImplementedError("Dividing by a unit is not allowed.")

	__rdiv__ = __rtruediv__

	def __add__(self, other: Union[float, "Unit"]) -> "Unit":
		if isinstance(other, Unit):
			return self.__class__.from_pt(float(self.as_pt()) + float(other.as_pt()))
		else:
			return self.__class__(super().__add__(other))

	__radd__ = __add__

	def __sub__(self, other: Union[float, "Unit"]) -> "Unit":
		if isinstance(other, Unit):
			return self.__class__.from_pt(float(self.as_pt()) - float(other.as_pt()))
		else:
			return self.__class__(super().__sub__(other))

	def __rsub__(self, other: Union[float, "Unit"]) -> "Unit":
		if isinstance(other, Unit):  # pragma: no cover (sub should be called instead)
			return self.__class__.from_pt(float(other.as_pt()) - float(self.as_pt()))
		else:
			return self.__class__(super().__rsub__(other))

	def as_pt(self) -> "Unit":
		"""
		Returns the unit in point.
		"""

		return Unit(float(_rounders(float(self) * self._in_pt, "0.000000")))

	@classmethod
	def from_pt(cls, value: float) -> "Unit":
		"""
		Construct a :class:`~.Unit` object from a value in point.

		:param value:
		"""

		return cls(value / cls._in_pt)

	def __call__(self, value: Union[SupportsFloat, str, bytes, bytearray] = 0.0) -> "Unit":
		"""
		Returns an instance of the :class:`Unit` with the given value.

		:param value:
		"""

		return self.__class__(value)


class Unitpt(Unit):
	"""
	Point.
	"""

	name = "pt"
	_in_pt = 1


class UnitInch(Unit):
	"""
	Inch.
	"""

	name = "inch"
	_in_pt = 72.0


class Unitcm(Unit):
	"""
	Centimetres.
	"""

	name = "cm"
	_in_pt = 28.3464566929


class Unitmm(Unit):
	"""
	Millimetres.
	"""

	name = "mm"
	_in_pt = 2.83464566929


class Unitum(Unit):
	"""
	Micrometres.
	"""

	name = "µm"
	_in_pt = 0.00283464566929


class Unitpc(Unit):
	"""
	Pica.
	"""

	name = "pc"
	_in_pt = 12.0


# Units
pt = Unitpt(1)  #: Point
inch = UnitInch(1)  #: Inch
cm = Unitcm(1)  #: Centimetre
mm = Unitmm(1)  #: Millimetre
um = Unitum(1)  #: Micrometre
pc = pica = Unitpc(1)  #: Pica



#!/usr/bin/env python
#
#  words.py
"""
Functions for working with (English) words.

.. versionadded:: 0.4.5
"""
#
#  Copyright © 2020-2022 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#
#  List of 10000 English words from https://github.com/first20hours/google-10000-english/
#  Derived from the Google Web Trillion Word Corpus,
#      as described by Thorsten Brants and Alex Franz,
#      and distributed by the Linguistic Data Consortium.
#
#  Subsets of this corpus distributed by Peter Novig.
#  Corpus editing and cleanup by Josh Kaufman.
#

# stdlib
import functools
import random
import re
import sys
from gettext import ngettext
from reprlib import recursive_repr
from string import ascii_lowercase, ascii_uppercase
from typing import Any, Dict, Iterable, List, NamedTuple, Optional, Tuple

# this package
import domdf_python_tools
from domdf_python_tools.compat import PYPY
from domdf_python_tools.doctools import prettify_docstrings

__all__ = [
		"greek_uppercase",
		"greek_lowercase",
		"get_words_list",
		"get_random_word",
		"make_font",
		"Font",
		"SERIF_BOLD_LETTERS",
		"SERIF_ITALIC_LETTERS",
		"SERIF_BOLD_ITALIC_LETTERS",
		"SANS_SERIF_LETTERS",
		"SANS_SERIF_BOLD_LETTERS",
		"SANS_SERIF_ITALIC_LETTERS",
		"SANS_SERIF_BOLD_ITALIC_LETTERS",
		"SCRIPT_LETTERS",
		"FRAKTUR_LETTERS",
		"MONOSPACE_LETTERS",
		"DOUBLESTRUCK_LETTERS",
		"alpha_sort",
		"as_text",
		"word_join",
		"TAB",
		"CR",
		"LF",
		"Plural",
		"PluralPhrase",
		"truncate_string",
		]

ascii_digits = "0123456789"
"""
ASCII numbers.

.. versionadded:: 0.7.0
"""

greek_uppercase = "ΑΒΓΔΕΖΗΘΙΚΛΜΝΞΟΠΡϴΣΤΥΦΧΨΩ"
"""
Uppercase Greek letters.

.. versionadded:: 0.7.0
"""

greek_lowercase = "αβγδεζηθικλμνξοπρςστυφχψω∂ϵϑϰϕϱϖ"
"""
Lowercase Greek letters.

.. versionadded:: 0.7.0
"""


@functools.lru_cache()
def get_words_list(min_length: int = 0, max_length: int = -1) -> List[str]:
	"""
	Returns the list of words, optionally only those whose length is between
	``min_length`` and ``max_length``.

	.. versionadded:: 0.4.5

	:param min_length: The minimum length of the words to return
	:param max_length: The maximum length of the words to return. A value of ``-1`` indicates no upper limit.
	:no-default max_length:

	:return: The list of words meeting the above specifiers.
	"""  # noqa: D400

	# this package
	from domdf_python_tools.compat import importlib_resources

	words: str = importlib_resources.read_text("domdf_python_tools", "google-10000-english-no-swears.txt")
	words_list: List[str] = words.splitlines()

	if min_length > 0 or max_length != -1:
		if max_length == -1:
			words_list = [word for word in words_list if min_length <= len(word)]
		else:
			words_list = [word for word in words_list if min_length <= len(word) <= max_length]

	return words_list


def get_random_word(min_length: int = 0, max_length: int = -1) -> str:
	"""
	Returns a random word, optionally only one whose length
	is between ``min_length`` and ``max_length``.

	.. versionadded:: 0.4.5

	:param min_length: The minimum length of the words to return
	:param max_length: The maximum length of the words to return. A value of ``-1`` indicates no upper limit.
	:no-default max_length:

	:return: A random word meeting the above specifiers.
	"""  # noqa: D400

	words_list = get_words_list(min_length=min_length, max_length=max_length)

	return random.choice(words_list)


# _default_unicode_sort_order: str = "".join(sorted(chr(i) for i in range(sys.maxunicode + 1)))


def alpha_sort(
		iterable: Iterable[str],
		alphabet: Iterable[str],  # = _default_unicode_sort_order
		reverse: bool = False,
		) -> List[str]:
	"""
	Sorts a list of strings using a custom alphabet.

	.. versionadded:: 0.7.0

	:param iterable: The strings to sort.
	:param alphabet: The custom alphabet to use for sorting.
	:param reverse:
	"""

	alphabet_ = list(alphabet)

	try:
		return sorted(iterable, key=lambda attr: [alphabet_.index(letter) for letter in attr], reverse=reverse)
	except ValueError as e:
		m = re.match(r"'(.*)' is not in list", str(e))
		if m:
			raise ValueError(f"The character {m.group(1)!r} was not found in the alphabet.") from None
		else:  # pragma: no cover
			raise e


class Font(Dict[str, str]):
	"""
	Represents a Unicode pseudo-font.

	Mapping of ASCII letters to their equivalents in the pseudo-font.

	Individual characters can be converted using the :meth:`Font.get <domdf_python_tools.words.Font.get>`
	method or the ``getitem`` syntax. Entire strings can be converted by calling the
	:class:`~domdf_python_tools.words.Font` object and passing the string as the first argument.
	"""

	def __getitem__(self, char: str) -> str:
		"""
		Returns the given character in this font.

		If the character is not found in this font the character is returned unchanged.

		:param char: The character to convert.
		"""

		char = str(char)

		if char not in self:
			return str(char)

		else:
			return str(super().__getitem__(char))

	def __call__(self, text: str) -> str:
		"""
		Returns the given text in this font.

		:param text:
		"""

		return ''.join(self[char] for char in text)

	def get(self, char: str, default: Optional[str] = None) -> str:  # type: ignore
		"""
		Returns the given character in this font.

		If the character is not found in this font the character is returned unchanged or,
		if a value for ``default`` is provided, that is returned instead.

		:param char: The character to convert.
		:param default: Optional default value.
		"""

		if char not in self and default is not None:
			return str(default)
		else:
			return self[char]


def make_font(
		uppers: Iterable[str],
		lowers: Iterable[str],
		digits: Optional[Iterable[str]] = None,
		greek_uppers: Optional[Iterable[str]] = None,
		greek_lowers: Optional[Iterable[str]] = None,
		) -> Font:
	"""
	Returns a dictionary mapping ASCII alphabetical characters and digits to the Unicode equivalents
	in a different pseudo-font.

	.. versionadded:: 0.7.0

	:param uppers: Iterable of uppercase letters (A-Z, 26 characters).
	:param lowers: Iterable of lowercase letters (a-z, 26 characters).
	:param digits: Optional iterable of digits (0-9).
	:param greek_uppers: Optional iterable of uppercase Greek letters (A-Ω, 25 characters).
	:param greek_lowers: Optional iterable of lowercase Greek letters (α-ϖ, 32 characters).
	"""  # noqa: D400

	font = Font({
			**dict(zip(ascii_uppercase, uppers)),
			**dict(zip(ascii_lowercase, lowers)),
			})

	if digits:
		font.update({char: unichar for char, unichar in zip(ascii_digits, digits)})

	if greek_uppers:
		font.update({char: unichar for char, unichar in zip(greek_uppercase, greek_uppers)})

	if greek_lowers:
		font.update({char: unichar for char, unichar in zip(greek_lowercase, greek_lowers)})

	return font


#: Bold Serif letters (uppercase)
SERIF_BOLD_UPPER = "𝐀𝐁𝐂𝐃𝐄𝐅𝐆𝐇𝐈𝐉𝐊𝐋𝐌𝐍𝐎𝐏𝐐𝐑𝐒𝐓𝐔𝐕𝐖𝐗𝐘𝐙"
#: Bold Serif letters (lowercase)
SERIF_BOLD_LOWER = "𝐚𝐛𝐜𝐝𝐞𝐟𝐠𝐡𝐢𝐣𝐤𝐥𝐦𝐧𝐨𝐩𝐪𝐫𝐬𝐭𝐮𝐯𝐰𝐱𝐲𝐳"
#: Bold Serif digits
SERIF_BOLD_DIGITS = "𝟎𝟏𝟐𝟑𝟒𝟓𝟔𝟕𝟖𝟗"
#: Bold Serif Greek letters (uppercase)
SERIF_BOLD_GREEK_UPPER = "𝚨𝚩𝚪𝚫𝚬𝚭𝚮𝚯𝚰𝚱𝚲𝚳𝚴𝚵𝚶𝚷𝚸𝚹𝚺𝚻𝚼𝚽𝚾𝚿𝛀"
#: Bold Serif Greek letters (lowercase)
SERIF_BOLD_GREEK_LOWER = "𝛂𝛃𝛄𝛅𝛆𝛇𝛈𝛉𝛊𝛋𝛌𝛍𝛎𝛏𝛐𝛑𝛒𝛓𝛔𝛕𝛖𝛗𝛘𝛙𝛚𝛛𝛜𝛝𝛞𝛟𝛠𝛡"

SERIF_BOLD_LETTERS = make_font(
		uppers=SERIF_BOLD_UPPER,
		lowers=SERIF_BOLD_LOWER,
		digits=SERIF_BOLD_DIGITS,
		greek_uppers=SERIF_BOLD_GREEK_UPPER,
		greek_lowers=SERIF_BOLD_GREEK_LOWER,
		)
"""
Bold Serif :class:`~domdf_python_tools.words.Font`.

This font includes numbers and Greek letters.

.. versionadded:: 0.7.0
"""

#: Italic Serif letters (uppercase)
SERIF_ITALIC_UPPER = "𝐴𝐵𝐶𝐷𝐸𝐹𝐺𝐻𝐼𝐽𝐾𝐿𝑀𝑁𝑂𝑃𝑄𝑅𝑆𝑇𝑈𝑉𝑊𝑋𝑌𝑍"
#: Italic Serif letters (lowercase)
SERIF_ITALIC_LOWER = "𝑎𝑏𝑐𝑑𝑒𝑓𝑔ℎ𝑖𝑗𝑘𝑙𝑚𝑛𝑜𝑝𝑞𝑟𝑠𝑡𝑢𝑣𝑤𝑥𝑦𝑧"
#: Italic Serif Greek letters (uppercase)
SERIF_ITALIC_GREEK_UPPER = "𝛢𝛣𝛤𝛥𝛦𝛧𝛨𝛩𝛪𝛫𝛬𝛭𝛮𝛯𝛰𝛱𝛲𝛳𝛴𝛵𝛶𝛷𝛸𝛹𝛺𝛻"
#: Italic Serif Greek letters (lowercase)
SERIF_ITALIC_GREEK_LOWER = "𝛼𝛽𝛾𝛿𝜀𝜁𝜂𝜃𝜄𝜅𝜆𝜇𝜈𝜉𝜊𝜋𝜌𝜍𝜎𝜏𝜐𝜑𝜒𝜓𝜔𝜕𝜖𝜗𝜘𝜙𝜚𝜛"

SERIF_ITALIC_LETTERS = make_font(
		uppers=SERIF_ITALIC_UPPER,
		lowers=SERIF_ITALIC_LOWER,
		greek_uppers=SERIF_ITALIC_GREEK_UPPER,
		greek_lowers=SERIF_ITALIC_GREEK_LOWER,
		)
"""
Italic Serif :class:`~domdf_python_tools.words.Font`.

This font includes Greek letters.

.. versionadded:: 0.7.0
"""

#: Bold and Italic Serif letters (uppercase)
SERIF_BOLD_ITALIC_UPPER = "𝑨𝑩𝑪𝑫𝑬𝑭𝑮𝑯𝑰𝑱𝑲𝑳𝑴𝑵𝑶𝑷𝑸𝑹𝑺𝑻𝑼𝑽𝑾𝑿𝒀𝒁"
#: Bold and Italic Serif letters (lowercase)
SERIF_BOLD_ITALIC_LOWER = "𝒂𝒃𝒄𝒅𝒆𝒇𝒈𝒉𝒊𝒋𝒌𝒍𝒎𝒏𝒐𝒑𝒒𝒓𝒔𝒕𝒖𝒗𝒘𝒙𝒚𝒛"
#: Bold and Italic Serif Greek letters (uppercase)
SERIF_BOLD_ITALIC_GREEK_UPPER = "𝜜𝜝𝜞𝜟𝜠𝜡𝜢𝜣𝜤𝜥𝜦𝜧𝜨𝜩𝜪𝜫𝜬𝜭𝜮𝜯𝜰𝜱𝜲𝜳𝜴𝜵"
#: Bold and Italic Serif Greek letters (lowercase)
SERIF_BOLD_ITALIC_GREEK_LOWER = "𝜶𝜷𝜸𝜹𝜺𝜻𝜼𝜽𝜾𝜿𝝀𝝁𝝂𝝃𝝄𝝅𝝆𝝇𝝈𝝉𝝊𝝋𝝌𝝍𝝎𝝏𝝐𝝑𝝒𝝓𝝔𝝕"

SERIF_BOLD_ITALIC_LETTERS = make_font(
		uppers=SERIF_BOLD_ITALIC_UPPER,
		lowers=SERIF_BOLD_ITALIC_LOWER,
		greek_uppers=SERIF_BOLD_ITALIC_GREEK_UPPER,
		greek_lowers=SERIF_BOLD_ITALIC_GREEK_LOWER,
		)
"""
Bold and Italic Serif :class:`~domdf_python_tools.words.Font`.

This font includes Greek letters.

.. versionadded:: 0.7.0
"""

#: Normal Sans-Serif letters (uppercase)
SANS_SERIF_UPPER = "𝖠𝖡𝖢𝖣𝖤𝖥𝖦𝖧𝖨𝖩𝖪𝖫𝖬𝖭𝖮𝖯𝖰𝖱𝖲𝖳𝖴𝖵𝖶𝖷𝖸𝖹"
#: Normal Sans-Serif letters (lowercase)
SANS_SERIF_LOWER = "𝖺𝖻𝖼𝖽𝖾𝖿𝗀𝗁𝗂𝗃𝗄𝗅𝗆𝗇𝗈𝗉𝗊𝗋𝗌𝗍𝗎𝗏𝗐𝗑𝗒𝗓"
#: Normal Sans-Serif digits
SANS_SERIF_DIGITS = "𝟢𝟣𝟤𝟥𝟦𝟧𝟨𝟩𝟪𝟫"

SANS_SERIF_LETTERS = make_font(
		uppers=SANS_SERIF_UPPER,
		lowers=SANS_SERIF_LOWER,
		digits=SANS_SERIF_DIGITS,
		)
"""
Normal Sans-Serif :class:`~domdf_python_tools.words.Font`.

This font includes numbers.

.. versionadded:: 0.7.0
"""

#: Bold Sans-Serif letters (uppercase)
SANS_SERIF_BOLD_UPPER = "𝗔𝗕𝗖𝗗𝗘𝗙𝗚𝗛𝗜𝗝𝗞𝗟𝗠𝗡𝗢𝗣𝗤𝗥𝗦𝗧𝗨𝗩𝗪𝗫𝗬𝗭"
#: Bold Sans-Serif letters (lowercase)
SANS_SERIF_BOLD_LOWER = "𝗮𝗯𝗰𝗱𝗲𝗳𝗴𝗵𝗶𝗷𝗸𝗹𝗺𝗻𝗼𝗽𝗾𝗿𝘀𝘁𝘂𝘃𝘄𝘅𝘆𝘇"
#: Bold Sans-Serif digits
SANS_SERIF_BOLD_DIGITS = "𝟬𝟭𝟮𝟯𝟰𝟱𝟲𝟳𝟴𝟵"

SANS_SERIF_BOLD_LETTERS = make_font(
		uppers=SANS_SERIF_BOLD_UPPER,
		lowers=SANS_SERIF_BOLD_LOWER,
		digits=SANS_SERIF_BOLD_DIGITS,
		)
"""
Bold Sans-Serif :class:`~domdf_python_tools.words.Font`.

This font includes numbers.

.. versionadded:: 0.7.0
"""

#: Italic Sans-Serif letters (uppercase)
SANS_SERIF_ITALIC_UPPER = "𝘈𝘉𝘊𝘋𝘌𝘍𝘎𝘏𝘐𝘑𝘒𝘓𝘔𝘕𝘖𝘗𝘘𝘙𝘚𝘛𝘜𝘝𝘞𝘟𝘠𝘡"
#: Italic Sans-Serif letters (lowercase)
SANS_SERIF_ITALIC_LOWER = "𝘢𝘣𝘤𝘥𝘦𝘧𝘨𝘩𝘪𝘫𝘬𝘭𝘮𝘯𝘰𝘱𝘲𝘳𝘴𝘵𝘶𝘷𝘸𝘹𝘺𝘻"

SANS_SERIF_ITALIC_LETTERS = make_font(
		uppers=SANS_SERIF_ITALIC_UPPER,
		lowers=SANS_SERIF_ITALIC_LOWER,
		)
"""
Italic Sans-Serif :class:`~domdf_python_tools.words.Font`.

.. versionadded:: 0.7.0
"""

#: Bold and Italic Sans-Serif letters (uppercase)
SANS_SERIF_BOLD_ITALIC_UPPER = "𝘼𝘽𝘾𝘿𝙀𝙁𝙂𝙃𝙄𝙅𝙆𝙇𝙈𝙉𝙊𝙋𝙌𝙍𝙎𝙏𝙐𝙑𝙒𝙓𝙔𝙕"
#: Bold and Italic Sans-Serif letters (lowercase)
SANS_SERIF_BOLD_ITALIC_LOWER = "𝙖𝙗𝙘𝙙𝙚𝙛𝙜𝙝𝙞𝙟𝙠𝙡𝙢𝙣𝙤𝙥𝙦𝙧𝙨𝙩𝙪𝙫𝙬𝙭𝙮𝙯"
#: Bold and Italic Sans-Serif letters (uppercase)
SANS_SERIF_BOLD_ITALIC_GREEK_UPPER = "𝞐𝞑𝞒𝞓𝞔𝞕𝞖𝞗𝞘𝞙𝞚𝞛𝞜𝞝𝞞𝞟𝞠𝞡𝞢𝞣𝞤𝞥𝞦𝞧𝞨𝞩"
#: Bold and Italic Sans-Serif letters (lowercase)
SANS_SERIF_BOLD_ITALIC_GREEK_LOWER = "𝞪𝞫𝞬𝞭𝞮𝞯𝞰𝞱𝞲𝞳𝞴𝞵𝞶𝞷𝞸𝞹𝞺𝞻𝞼𝞽𝞾𝞿𝟀𝟁𝟂𝟃𝟄𝟅𝟆𝟇𝟈𝟉"

SANS_SERIF_BOLD_ITALIC_LETTERS = make_font(
		uppers=SANS_SERIF_BOLD_ITALIC_UPPER,
		lowers=SANS_SERIF_BOLD_ITALIC_LOWER,
		greek_uppers=SANS_SERIF_BOLD_ITALIC_GREEK_UPPER,
		greek_lowers=SANS_SERIF_BOLD_ITALIC_GREEK_LOWER,
		)
"""
Bold and Italic Sans-Serif :class:`~domdf_python_tools.words.Font`.

This font includes Greek letters.

.. versionadded:: 0.7.0
"""

#: Script letters (uppercase)
SCRIPT_UPPER = "𝓐𝓑𝓒𝓓𝓔𝓕𝓖𝓗𝓘𝓙𝓚𝓛𝓜𝓝𝓞𝓟𝓠𝓡𝓢𝓣𝓤𝓥𝓦𝓧𝓨𝓩"
#: Script letters (lowercase)
SCRIPT_LOWER = "𝓪𝓫𝓬𝓭𝓮𝓯𝓰𝓱𝓲𝓳𝓴𝓵𝓶𝓷𝓸𝓹𝓺𝓻𝓼𝓽𝓾𝓿𝔀𝔁𝔂𝔃"

SCRIPT_LETTERS = make_font(SCRIPT_UPPER, SCRIPT_LOWER)
"""
Script :class:`~domdf_python_tools.words.Font`.

.. versionadded:: 0.7.0
"""

#: Fraktur letters (uppercase)
FRAKTUR_UPPER = "𝕬𝕭𝕮𝕯𝕰𝕱𝕲𝕳𝕴𝕵𝕶𝕷𝕸𝕹𝕺𝕻𝕼𝕽𝕾𝕿𝖀𝖁𝖂𝖃𝖄𝖅"
#: Fraktur letters (lowercase)
FRAKTUR_LOWER = "𝖆𝖇𝖈𝖉𝖊𝖋𝖌𝖍𝖎𝖏𝖐𝖑𝖒𝖓𝖔𝖕𝖖𝖗𝖘𝖙𝖚𝖛𝖜𝖝𝖞𝖟"

FRAKTUR_LETTERS = make_font(FRAKTUR_UPPER, FRAKTUR_LOWER)
"""
Fraktur :class:`~domdf_python_tools.words.Font`.

.. versionadded:: 0.7.0
"""

#: Monospace letters (uppercase)
MONOSPACE_UPPER = "𝙰𝙱𝙲𝙳𝙴𝙵𝙶𝙷𝙸𝙹𝙺𝙻𝙼𝙽𝙾𝙿𝚀𝚁𝚂𝚃𝚄𝚅𝚆𝚇𝚈𝚉"
#: Monospace letters (lowercase)
MONOSPACE_LOWER = "𝚊𝚋𝚌𝚍𝚎𝚏𝚐𝚑𝚒𝚓𝚔𝚕𝚖𝚗𝚘𝚙𝚚𝚛𝚜𝚝𝚞𝚟𝚠𝚡𝚢𝚣"
#: Monospace digits
MONOSPACE_DIGITS = "𝟶𝟷𝟸𝟹𝟺𝟻𝟼𝟽𝟾𝟿"

MONOSPACE_LETTERS = make_font(MONOSPACE_UPPER, MONOSPACE_LOWER, MONOSPACE_DIGITS)
"""
Monospace :class:`~domdf_python_tools.words.Font`.

This font includes numbers.

.. versionadded:: 0.7.0
"""

#: Doublestruck letters (uppercase)
DOUBLESTRUCK_UPPER = "𝔸𝔹ℂ𝔻𝔼𝔽𝔾ℍ𝕀𝕁𝕂𝕃𝕄ℕ𝕆ℙℚℝ𝕊𝕋𝕌𝕍𝕎𝕏𝕐ℤ"
#: Doublestruck letters (lowercase)
DOUBLESTRUCK_LOWER = "𝕒𝕓𝕔𝕕𝕖𝕗𝕘𝕙𝕚𝕛𝕜𝕝𝕞𝕟𝕠𝕡𝕢𝕣𝕤𝕥𝕦𝕧𝕨𝕩𝕪𝕫"
#: Doublestruck digits
DOUBLESTRUCK_DIGITS = "𝟘𝟙𝟚𝟛𝟜𝟝𝟞𝟟𝟠𝟡"

DOUBLESTRUCK_LETTERS = make_font(DOUBLESTRUCK_UPPER, DOUBLESTRUCK_LOWER, DOUBLESTRUCK_DIGITS)
"""
Doublestruck :class:`~domdf_python_tools.words.Font`.

This font includes numbers.

.. versionadded:: 0.7.0
"""


def as_text(value: Any) -> str:
	"""
	Convert the given value to a string. :py:obj:`None` is converted to ``''``.

	:param value: The value to convert to a string.

	:rtype:

	.. versionchanged:: 0.8.0

		Moved from :mod:`domdf_python_tools.utils`.
	"""

	if value is None:
		return ''

	return str(value)


def word_join(
		iterable: Iterable[str],
		use_repr: bool = False,
		oxford: bool = False,
		delimiter: str = ',',
		connective: str = "and",
		) -> str:
	"""
	Join the given list of strings in a natural manner, with 'and' to join the last two elements.

	:param iterable:
	:param use_repr: Whether to join the ``repr`` of each object.
	:param oxford: Whether to use an oxford comma when joining the last two elements.
	:default oxford: :py:obj:`False`. Always :py:obj:`False` if there are fewer than three elements
	:param delimiter: A single character to use between the words.
	:param connective: The connective to join the final two words with.

	:rtype:

	.. versionchanged:: 0.11.0

		Added ``delimiter`` and ``connective`` arguments.
	"""

	delimiter = f"{delimiter} "

	if use_repr:
		words = [repr(w) for w in iterable]
	else:
		words = list(iterable)

	if len(words) == 0:
		return ''
	elif len(words) == 1:
		return words[0]
	elif len(words) == 2:
		return f" {connective} ".join(words)
	else:
		if oxford:
			return delimiter.join(words[:-1]) + f"{delimiter}{connective} {words[-1]}"
		else:
			return delimiter.join(words[:-1]) + f" {connective} {words[-1]}"


TAB = '\t'
"""
A literal ``TAB`` (``\\t``) character for use in f-strings.

.. versionadded:: 1.3.0
"""

CR = '\r'
"""
The carriage return character (``\\r``) for use in f-strings.

.. versionadded:: 1.3.0
"""

LF = '\n'
"""
The newline character (``\\n``) for use in f-strings.

.. versionadded:: 1.3.0
"""

_docs = domdf_python_tools.__docs


@prettify_docstrings
class Plural(functools.partial):
	"""
	Represents a word as its singular and plural.

	.. versionadded:: 2.0.0

	:param singular: The singular form of the word.
	:param plural: The plural form of the word.

	.. code-block:: python

		>>> cow = Plural("cow", "cows")
		>>> n = 1
		>>> print(f"The farmer has {n} {cow(n)}.")
		The farmer has 1 cow.
		>>> n = 2
		>>> print(f"The farmer has {n} {cow(n)}.")
		The farmer has 2 cows.
		>>> n = 3
		>>> print(f"The farmer has {n} {cow(n)}.")
		The farmer has 3 cows.
	"""

	if _docs:  # pragma: no cover

		def __init__(self, singular: str, plural: str):
			pass

		def __call__(self, n: int) -> str:  # type: ignore
			"""
			Returns either the singular or plural form of the word depending on the value of ``n``.

			:param n:
			"""

	# if PYPY:  # pragma: no cover (!PyPy)
	if PYPY and sys.version_info < (3, 9):  # pragma: no cover (!PyPy)

		def __init__(self, singular: str, plural: str):
			super().__init__(ngettext, singular, plural)  # type: ignore[call-arg]
	else:  # pragma: no cover (!CPython)

		def __new__(cls, singular: str, plural: str):  # noqa: D102
			return functools.partial.__new__(cls, ngettext, singular, plural)

	@recursive_repr()
	def __repr__(self):
		qualname = type(self).__qualname__
		args: List[str] = []
		args.extend(repr(x) for x in self.args)
		args.extend(f"{k}={v!r}" for (k, v) in self.keywords.items())
		return f"{qualname}({', '.join(args)})"


@prettify_docstrings
class PluralPhrase(NamedTuple):
	"""
	Represents a phrase which varies depending on a numerical count.

	.. versionadded:: 3.3.0

	:param template: The phrase template.
	:param words: The words to insert into the template.

	For example, consider the phase::

		The proposed changes are to ...

	The "phrase template" would be:

	.. code-block:: python

		"The proposed {} {} to ..."

	and the two words to insert are:

	.. code-block:: python

		Plural("change", "changes")
		Plural("is", "are")

	The phrase is constructed as follows:

	.. code-block:: python

		>>> phrase = PluralPhrase(
		...     "The proposed {} {} to ...",
		...     (Plural("change", "changes"), Plural("is", "are"))
		... )
		>>> phrase(1)
		'The proposed change is to ...'
		>>> phrase(2)
		'The proposed changes are to ...'

	The phrase template can use any `valid syntax`_ for :meth:`str.format`,
	except for keyword arguments. The exception if the keyword ``n``,
	which is replaced with the count (e.g. ``2``) passed in when the phrase is constructed.
	For example:

	.. code-block:: python

		>>> phrase2 = PluralPhrase("The farmer has {n} {0}.", (Plural("cow", "cows"), ))
		>>> phrase2(2)
		'The farmer has 2 cows.'

	.. _valid syntax: https://docs.python.org/3/library/string.html#formatstrings

	"""

	template: str
	words: Tuple[Plural, ...]

	def __call__(self, n: int) -> str:  # noqa: TYP004  # TODO
		"""
		Construct the phrase based on the value of ``n``.

		:param n:
		"""

		plural_words = [x(n) for x in self.words]
		return self.template.format(*plural_words, n=n)


@functools.lru_cache()
def _slice_end(max_length: int, ending: str = "...") -> slice:
	slice_end = max_length - len(ending)
	return slice(slice_end)


def truncate_string(string: str, max_length: int, ending: str = "...") -> str:
	"""
	Truncate a string to ``max_length`` characters, and put ``ending`` on the end.

	The truncated string is further truncated by the length of ``ending`` so the returned string is no more then ``max_length``.

	.. versionadded:: 3.3.0

	:param string:
	:param max_length:
	:param ending:
	"""

	string_length = len(string)

	if string_length > max_length:
		return string[_slice_end(max_length, ending)] + ending
	else:
		return string



#  !/usr/bin/env python
#
#  sizes.py
"""
Common pagesizes in point/pt.

.. TODO:: finish the list of the page sizes

.. |iso216| replace:: `ISO 216 <https://en.wikipedia.org/wiki/ISO_216>`__

Each pagesize is an instance of :class:`domdf_python_tools.pagesizes.PageSize`.
The following sizes are available:
"""
#
#  Copyright © 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Based on reportlab.lib.pagesizes and reportlab.lib.units
#    www.reportlab.co.uk
#    Copyright ReportLab Europe Ltd. 2000-2017
#    Copyright (c) 2000-2018, ReportLab Inc.
#    All rights reserved.
#    Licensed under the BSD License
#
#  Includes data from en.wikipedia.org.
#  Licensed under the Creative Commons Attribution-ShareAlike License
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

__all__ = [
		"_4A0",
		"_2A0",
		"A0",
		"A1",
		"A2",
		"A3",
		"A4",
		"A5",
		"A6",
		"A7",
		"A8",
		"A9",
		"A10",
		"B0",
		"B1",
		"B2",
		"B3",
		"B4",
		"B5",
		"B6",
		"B7",
		"B8",
		"B9",
		"B10",
		"C0",
		"C1",
		"C2",
		"C3",
		"C4",
		"C5",
		"C6",
		"C7",
		"C8",
		"C9",
		"C10",
		"A2EXTRA",
		"A3EXTRA",
		"A3SUPER",
		"SUPERA3",
		"A4EXTRA",
		"A4SUPER",
		"SUPERA4",
		"A4LONG",
		"A5EXTRA",
		"SOB5EXTRA",
		"LETTER",
		"LEGAL",
		"TABLOID",
		"ELEVENSEVENTEEN",
		"JUNIOR_LEGAL",
		"HALF_LETTER",
		"GOV_LETTER",
		"GOV_LEGAL",
		"LEDGER",
		"EMPEROR",
		"QUAD_ROYAL",
		"QUAD_DEMY",
		"ANTIQUARIAN",
		"GRAND_EAGLE",
		"DOUBLE_ELEPHANT",
		"ATLAS",
		"DOUBLE_ROYAL",
		"COLOMBIER",
		"DOUBLE_DEMY_US",
		"DOUBLE_DEMY",
		"DOUBLE_DEMY_UK",
		"IMPERIAL",
		"DOUBLE_LARGE_POST",
		"ELEPHANT",
		"PRINCESS",
		"CARTRIDGE",
		"ROYAL",
		"SHEET",
		"HALF_POST",
		"DOUBLE_POST",
		"SUPER_ROYAL",
		"BROADSHEET",
		"MEDIUM_US",
		"MEDIUM_UK",
		"DEMY",
		"COPY_DRAUGHT",
		"LARGE_POST_US",
		"LARGE_POST_UK",
		"POST_US",
		"POST_UK",
		"CROWN",
		"PINCHED_POST",
		"FOOLSCAP_US",
		"FOOLSCAP_UK",
		"SMALL_FOOLSCAP",
		"BRIEF",
		"POTT",
		"QUARTO_US",
		"EXECUTIVE",
		"MONARCH",
		"FOLIO",
		"FOOLSCAP_FOLIO",
		"QUARTO",
		"QUARTO_UK",
		"KINGS",
		"DUKES",
		"ID_1",
		"ID_2",
		"ID_3",
		"ID_000",
		]

# this package
from .classes import Size_inch, Size_mm

# ISO 216 standard paper sizes; see eg https://en.wikipedia.org/wiki/ISO_216
# also http://www.printernational.org/iso-paper-sizes.php

_4A0 = Size_mm(1682, 2378).to_pt()  #: |iso216| 4A0 Paper
_2A0 = Size_mm(1189, 1682).to_pt()  #: |iso216| 2A0 Paper
A0 = Size_mm(841, 1189).to_pt()  #: |iso216| A0 Paper
A1 = Size_mm(594, 841).to_pt()  #: |iso216| A1 Paper
A2 = Size_mm(420, 594).to_pt()  #: |iso216| A2 Paper
A3 = Size_mm(297, 420).to_pt()  #: |iso216| A3 Paper
A4 = Size_mm(210, 297).to_pt()  #: |iso216| A3 Paper
A5 = Size_mm(148, 210).to_pt()  #: |iso216| A5 Paper
A6 = Size_mm(105, 148).to_pt()  #: |iso216| A6 Paper
A7 = Size_mm(74, 105).to_pt()  #: |iso216| A7 Paper
A8 = Size_mm(52, 74).to_pt()  #: |iso216| A8 Paper
A9 = Size_mm(37, 52).to_pt()  #: |iso216| A0 Paper
A10 = Size_mm(26, 37).to_pt()  #: |iso216| A10 Paper

# _W, _H = (21 * cm, 29.7 * cm)
# A6 = (_W * .5, _H * .5)
# A5 = (_H * .5, _W)
# A4 = (_W, _H)
# A3 = (_H, _W * 2)
# A2 = (_W * 2, _H * 2)
# A1 = (_H * 2, _W * 4)
# A0 = (_W * 4, _H * 4)

B0 = Size_mm(1000, 1414)  #: |iso216| B0 Paper
B1 = Size_mm(707, 1000)  #: |iso216| B1 Paper
B2 = Size_mm(500, 707)  #: |iso216| B2 Paper
B3 = Size_mm(353, 500)  #: |iso216| B3 Paper
B4 = Size_mm(250, 353)  #: |iso216| B4 Paper
B5 = Size_mm(176, 250)  #: |iso216| B5 Paper
B6 = Size_mm(125, 176)  #: |iso216| B6 Paper
B7 = Size_mm(88, 125)  #: |iso216| B7 Paper
B8 = Size_mm(62, 88)  #: |iso216| B8 Paper
B9 = Size_mm(44, 62)  #: |iso216| B9 Paper
B10 = Size_mm(31, 44)  #: |iso216| B10 Paper

# _BW, _BH = (25 * cm, 35.3 * cm)
# B6 = (_BW * .5, _BH * .5)
# B5 = (_BH * .5, _BW)
# B4 = (_BW, _BH)
# B3 = (_BH * 2, _BW)
# B2 = (_BW * 2, _BH * 2)
# B1 = (_BH * 4, _BW * 2)
# B0 = (_BW * 4, _BH * 4)

C0 = Size_mm(917, 1297)  #: |iso216| C0 Paper
C1 = Size_mm(648, 917)  #: |iso216| C1 Paper
C2 = Size_mm(458, 648)  #: |iso216| C2 Paper
C3 = Size_mm(324, 458)  #: |iso216| C3 Paper
C4 = Size_mm(229, 324)  #: |iso216| C4 Paper
C5 = Size_mm(162, 229)  #: |iso216| C5 Paper
C6 = Size_mm(114, 162)  #: |iso216| C6 Paper
C7 = Size_mm(81, 114)  #: |iso216| C7 Paper
C8 = Size_mm(57, 81)  #: |iso216| C8 Paper
C9 = Size_mm(40, 57)  #: |iso216| C9 Paper
C10 = Size_mm(28, 40)  #: |iso216| C10 Paper

A2EXTRA = Size_mm(445, 619)  #: A2 Extra Paper
A3EXTRA = Size_mm(322, 445)  #: A3 Extra Paper
A3SUPER = Size_mm(305, 508)  #: A3 Super Paper (different to (Super A3)
SUPERA3 = Size_mm(305, 487)  #: Super A3 Paper (different to A3 Super)
A4EXTRA = Size_mm(235, 322)  #: A4 Extra Paper
A4SUPER = Size_mm(229, 322)  #: A4 Super Paper (different to (Super A4)
SUPERA4 = Size_mm(227, 356)  #: Super A3 Paper (different to A4 Super)
A4LONG = Size_mm(210, 348)  #: A4 Long Paper
A5EXTRA = Size_mm(173, 235)  #: A4 Extra Paper
SOB5EXTRA = Size_mm(202, 276)  #: SO B5 Extra Paper

# American paper sizes
LETTER = Size_inch(8.5, 11).to_pt()  #: North American "Letter" Paper
LEGAL = Size_inch(8.5, 14).to_pt()  #: North American "Legal" Paper
TABLOID = ELEVENSEVENTEEN = Size_inch(11, 17).to_pt()

# From https://en.wikipedia.org/wiki/Paper_size
JUNIOR_LEGAL = Size_inch(5, 8).to_pt()  #: Junior Legal
HALF_LETTER = Size_inch(5.5, 8).to_pt()  #: Half Letter
GOV_LETTER = Size_inch(8, 10.5).to_pt()  #: Government Letter
GOV_LEGAL = Size_inch(8.5, 13).to_pt()  #: Government Legal
LEDGER = Size_inch(17, 11).to_pt()  #: Ledger
EMPEROR = Size_inch(48, 72).to_pt()  #: Emperor
QUAD_ROYAL = Size_inch(40, 50).to_pt()  #: Quad Royal
QUAD_DEMY = Size_inch(35, 40).to_pt()  #:
ANTIQUARIAN = Size_inch(31, 53).to_pt()  #:
GRAND_EAGLE = Size_inch(28.75, 42).to_pt()  #:
DOUBLE_ELEPHANT = Size_inch(26.75, 40).to_pt()  #:
ATLAS = Size_inch(26, 34).to_pt()  #:
DOUBLE_ROYAL = Size_inch(25, 40).to_pt()  #:
COLOMBIER = Size_inch(23.5, 34.5).to_pt()  #:
DOUBLE_DEMY_US = Size_inch(22.5, 35.5).to_pt()  #:
DOUBLE_DEMY = DOUBLE_DEMY_UK = Size_inch(22.5, 35).to_pt()  #:
IMPERIAL = Size_inch(22, 30).to_pt()  #:
DOUBLE_LARGE_POST = Size_inch(21, 33).to_pt()  #:
ELEPHANT = Size_inch(23, 28).to_pt()  #:
PRINCESS = Size_inch(22.5, 28).to_pt()  #:
CARTRIDGE = Size_inch(21, 26).to_pt()  #:
ROYAL = Size_inch(20, 25).to_pt()  #:
SHEET = HALF_POST = Size_inch(19.5, 23.5).to_pt()  #:
DOUBLE_POST = Size_inch(19, 30.5).to_pt()  #:
SUPER_ROYAL = Size_inch(19, 27).to_pt()  #:
BROADSHEET = Size_inch(18, 24).to_pt()  #:
MEDIUM_US = Size_inch(17.5, 23).to_pt()  #:
MEDIUM_UK = Size_inch(18, 23).to_pt()  #:
DEMY = Size_inch(17.5, 22.5).to_pt()  #:
COPY_DRAUGHT = Size_inch(16, 20).to_pt()  #:
LARGE_POST_US = Size_inch(15.5, 20).to_pt()  #:
LARGE_POST_UK = Size_inch(16.5, 21).to_pt()  #:
POST_US = Size_inch(15.5, 19.35).to_pt()  #:
POST_UK = Size_inch(15.5, 19.5).to_pt()  #:
CROWN = Size_inch(15, 20).to_pt()  #:
PINCHED_POST = Size_inch(14.75, 18.5).to_pt()  #:
FOOLSCAP_US = Size_inch(13.5, 17).to_pt()  #:
FOOLSCAP_UK = Size_inch(13, 18).to_pt()  #:
SMALL_FOOLSCAP = Size_inch(13.35, 16.5).to_pt()  #:
BRIEF = Size_inch(13.5, 16).to_pt()  #:
POTT = Size_inch(12.5, 15).to_pt()  #:
QUARTO_US = Size_inch(9, 11).to_pt()  #:
EXECUTIVE = MONARCH = Size_inch(7.35, 10.5).to_pt()  #:
FOLIO = FOOLSCAP_FOLIO = Size_inch(8, 13).to_pt()  #:
QUARTO = QUARTO_UK = Size_inch(8, 10).to_pt()  #:
# IMPERIAL = Size_inch(7*inch, 9*inch).to_pt()  there are two of these?
KINGS = Size_inch(6.5, 8).to_pt()  #:
DUKES = Size_inch(5.5, 7).to_pt()  #:

# https://en.wikipedia.org/wiki/ISO/IEC_7810
ID_1 = Size_mm(85.60, 53.98).to_pt()  #: Most banking cards and ID cards
ID_2 = Size_mm(105, 74).to_pt()  #: French and other ID cards; Visas
ID_3 = Size_mm(125, 88).to_pt()  #: US government ID cards
ID_000 = Size_mm(25, 15).to_pt()  #: SIM cards



#  !/usr/bin/env python
#
#  bases.py
"""
Useful base classes.
"""
#
#  Copyright © 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#
#  UserList based on CPython.
#  Licensed under the Python Software Foundation License Version 2.
#  Copyright © 2001-2020 Python Software Foundation. All rights reserved.
#  Copyright © 2000 BeOpen.com. All rights reserved.
#  Copyright © 1995-2000 Corporation for National Research Initiatives. All rights reserved.
#  Copyright © 1991-1995 Stichting Mathematisch Centrum. All rights reserved.
#

# stdlib
from abc import abstractmethod
from numbers import Real
from pprint import pformat
from typing import (
		Any,
		Dict,
		Iterable,
		Iterator,
		List,
		MutableSequence,
		Optional,
		SupportsFloat,
		Tuple,
		Type,
		TypeVar,
		Union,
		overload
		)

# this package
from domdf_python_tools._is_match import is_match_with
from domdf_python_tools.doctools import prettify_docstrings
from domdf_python_tools.typing import SupportsIndex

__all__ = [
		"Dictable",
		"NamedList",
		"namedlist",
		"UserList",
		"UserFloat",
		"Lineup",
		"_V",
		"_LU",
		"_T",
		"_S",
		"_F",
		]

_F = TypeVar("_F", bound="UserFloat")
_LU = TypeVar("_LU", bound="Lineup")
_S = TypeVar("_S", bound="UserList")
_T = TypeVar("_T")
_V = TypeVar("_V")


@prettify_docstrings
class Dictable(Iterable[Tuple[str, _V]]):
	"""
	The basic structure of a class that can be converted into a dictionary.
	"""

	@abstractmethod
	def __init__(self, *args, **kwargs):
		pass

	def __repr__(self) -> str:
		return super().__repr__()

	def __str__(self) -> str:
		return self.__repr__()

	def __iter__(self) -> Iterator[Tuple[str, _V]]:
		"""
		Iterate over the attributes of the class.
		"""

		yield from self.__dict__.items()

	def __getstate__(self) -> Dict[str, _V]:
		return self.__dict__

	def __setstate__(self, state):
		self.__init__(**state)  # type: ignore[misc]

	def __copy__(self):
		return self.__class__(**self.__dict__)

	def __deepcopy__(self, memodict={}):
		return self.__copy__()

	@property
	@abstractmethod
	def __dict__(self):
		return dict()  # pragma: no cover (abc)

	def __eq__(self, other) -> bool:
		if isinstance(other, self.__class__):
			return is_match_with(other.__dict__, self.__dict__)

		return NotImplemented


@prettify_docstrings
class UserList(MutableSequence[_T]):
	"""
	Typed version of :class:`collections.UserList`.

	Class that simulates a list. The instance’s contents are kept in a regular list,
	which is accessible via the :attr:`~.UserList.data` attribute of :class:`~.UserList` instances.
	The instance’s contents are initially set to a copy of list, defaulting to the empty list ``[]``.

	.. versionadded:: 0.10.0

	:param initlist: The initial values to populate the :class:`~.UserList` with.
	:default initlist: ``[]``

	.. latex:clearpage::

	.. admonition:: Subclassing requirements

		Subclasses of :class:`~.UserList` are expected to offer a constructor which can be called with
		either no arguments or one argument. List operations which return a new sequence
		attempt to create an instance of the actual implementation class. To do so,
		it assumes that the constructor can be called with a single parameter, which is a
		sequence object used as a data source.

		If a derived class does not wish to comply with this requirement, all of the special
		methods supported by this class will need to be overridden; please consult the
		sources for information about the methods which need to be provided in that case.
	"""

	#: A real list object used to store the contents of the :class:`~domdf_python_tools.bases.UserList`.
	data: List[_T]

	def __init__(self, initlist: Optional[Iterable[_T]] = None):
		self.data = []
		if initlist is not None:
			# XXX should this accept an arbitrary sequence?
			if type(initlist) is type(self.data):  # noqa: E721
				self.data[:] = initlist
			elif isinstance(initlist, UserList):
				self.data[:] = initlist.data[:]
			else:
				self.data = list(initlist)

	def __repr__(self) -> str:
		return repr(self.data)

	def __lt__(self, other: object) -> bool:
		return self.data < self.__cast(other)

	def __le__(self, other: object) -> bool:
		return self.data <= self.__cast(other)

	def __eq__(self, other: object) -> bool:
		return self.data == self.__cast(other)

	def __gt__(self, other: object) -> bool:
		return self.data > self.__cast(other)

	def __ge__(self, other: object) -> bool:
		return self.data >= self.__cast(other)

	@staticmethod
	def __cast(other):
		return other.data if isinstance(other, UserList) else other

	def __contains__(self, item: object) -> bool:
		return item in self.data

	def __len__(self) -> int:
		return len(self.data)

	def __iter__(self) -> Iterator[_T]:
		yield from self.data

	@overload
	def __getitem__(self, i: int) -> _T: ...

	@overload
	def __getitem__(self, i: slice) -> MutableSequence[_T]: ...

	def __getitem__(self, i: Union[int, slice]) -> Union[_T, MutableSequence[_T]]:
		if isinstance(i, slice):
			return self.__class__(self.data[i])
		else:
			return self.data[i]

	@overload
	def __setitem__(self, i: int, o: _T) -> None: ...

	@overload
	def __setitem__(self, i: slice, o: Iterable[_T]) -> None: ...

	def __setitem__(self, i: Union[int, slice], item: Union[_T, Iterable[_T]]) -> None:
		self.data[i] = item  # type: ignore[index, assignment]

	def __delitem__(self, i: Union[int, slice]):
		del self.data[i]

	def __add__(self: _S, other: Iterable[_T]) -> _S:
		if isinstance(other, UserList):
			return self.__class__(self.data + other.data)
		elif isinstance(other, type(self.data)):
			return self.__class__(self.data + other)
		return self.__class__(self.data + list(other))

	def __radd__(self, other):
		if isinstance(other, UserList):
			return self.__class__(other.data + self.data)
		elif isinstance(other, type(self.data)):
			return self.__class__(other + self.data)
		return self.__class__(list(other) + self.data)

	def __iadd__(self: _S, other: Iterable[_T]) -> _S:
		if isinstance(other, UserList):
			self.data += other.data
		elif isinstance(other, type(self.data)):
			self.data += other
		else:
			self.data += list(other)
		return self

	def __mul__(self: _S, n: int) -> _S:
		return self.__class__(self.data * n)

	__rmul__ = __mul__

	def __imul__(self: _S, n: int) -> _S:
		self.data *= n
		return self

	def __copy__(self):
		inst = self.__class__.__new__(self.__class__)
		inst.__dict__.update(self.__dict__)
		# Create a copy and avoid triggering descriptors
		inst.__dict__["data"] = self.__dict__["data"][:]
		return inst

	def append(self, item: _T) -> None:
		"""
		Append ``item`` to the end of the :class:`~.domdf_python_tools.bases.UserList`.
		"""

		self.data.append(item)

	def insert(self, i: int, item: _T) -> None:
		"""
		Insert ``item`` at position ``i`` in the :class:`~.domdf_python_tools.bases.UserList`.
		"""

		self.data.insert(i, item)

	def pop(self, i: int = -1) -> _T:
		"""
		Removes and returns the item at index ``i``.

		:raises IndexError: if list is empty or index is out of range.
		"""

		return self.data.pop(i)

	def remove(self, item: _T) -> None:
		"""
		Removes the first occurrence of ``item`` from the list.

		:param item:

		:rtype:

		:raises ValueError: if the item is not present.

		.. latex:clearpage::
		"""

		self.data.remove(item)

	def clear(self) -> None:
		"""
		Remove all items from the :class:`~.domdf_python_tools.bases.UserList`.
		"""

		self.data.clear()

	def copy(self: _S) -> _S:
		"""
		Returns a copy of the :class:`~.domdf_python_tools.bases.UserList`.
		"""

		return self.__class__(self)

	def count(self, item: _T) -> int:
		"""
		Returns the number of occurrences of ``item`` in the :class:`~.domdf_python_tools.bases.UserList`.
		"""

		return self.data.count(item)

	def index(self, item: _T, *args: Any) -> int:
		"""
		Returns the index of the fist element matching ``item``.

		:param item:
		:param args:

		:raises ValueError: if the item is not present.
		"""

		return self.data.index(item, *args)

	def reverse(self) -> None:
		"""
		Reverse the list in place.
		"""

		self.data.reverse()

	def sort(self, *, key=None, reverse: bool = False) -> None:
		"""
		Sort the list in ascending order and return :py:obj:`None`.

		The sort is in-place (i.e. the list itself is modified) and stable (i.e. the
		order of two equal elements is maintained).

		If a key function is given, apply it once to each list item and sort them,
		ascending or descending, according to their function values.

		The reverse flag can be set to sort in descending order.
		"""

		self.data.sort(key=key, reverse=reverse)

	def extend(self, other: Iterable[_T]) -> None:
		"""
		Extend the :class:`~.domdf_python_tools.bases.NamedList` by appending elements from ``other``.

		:param other:
		"""

		if isinstance(other, UserList):
			self.data.extend(other.data)
		else:
			self.data.extend(other)


@prettify_docstrings
class UserFloat(Real):
	"""
	Class which simulates a float.

	.. versionadded:: 1.6.0

	:param value: The values to initialise the :class:`~domdf_python_tools.bases.UserFloat` with.
	"""

	def __init__(self, value: Union[SupportsFloat, SupportsIndex, str, bytes, bytearray] = 0.0):
		self._value = (float(value), )

	def as_integer_ratio(self) -> Tuple[int, int]:
		"""
		Returns the float as a fraction.
		"""

		return float(self).as_integer_ratio()

	def hex(self) -> str:  # noqa: A003  # pylint: disable=redefined-builtin
		"""
		Returns the hexadecimal (base 16) representation of the float.
		"""

		return float(self).hex()

	def is_integer(self) -> bool:
		"""
		Returns whether the float is an integer.
		"""

		return float(self).is_integer()

	@classmethod
	def fromhex(cls: Type[_F], string: str) -> _F:
		"""
		Create a floating-point number from a hexadecimal string.

		:param string:
		"""

		return cls(float.fromhex(string))

	def __add__(self: _F, other: float) -> _F:
		return self.__class__(float(self).__add__(other))

	def __sub__(self: _F, other: float) -> _F:
		return self.__class__(float(self).__sub__(other))

	def __mul__(self: _F, other: float) -> _F:
		return self.__class__(float(self).__mul__(other))

	def __floordiv__(self: _F, other: float) -> _F:  # type: ignore[override]
		return self.__class__(float(self).__floordiv__(other))

	def __truediv__(self: _F, other: float) -> _F:
		return self.__class__(float(self).__truediv__(other))

	def __mod__(self: _F, other: float) -> _F:
		return self.__class__(float(self).__mod__(other))

	def __divmod__(self: _F, other: float) -> Tuple[_F, _F]:
		return tuple(self.__class__(x) for x in float(self).__divmod__(other))  # type: ignore[return-value]

	def __pow__(self: _F, other: float, mod=None) -> _F:
		return self.__class__(float(self).__pow__(other, mod))

	def __radd__(self: _F, other: float) -> _F:
		return self.__class__(float(self).__radd__(other))

	def __rsub__(self: _F, other: float) -> _F:
		return self.__class__(float(self).__rsub__(other))

	def __rmul__(self: _F, other: float) -> _F:
		return self.__class__(float(self).__rmul__(other))

	def __rfloordiv__(self: _F, other: float) -> _F:  # type: ignore[override]
		return self.__class__(float(self).__rfloordiv__(other))

	def __rtruediv__(self: _F, other: float) -> _F:
		return self.__class__(float(self).__rtruediv__(other))

	def __rmod__(self: _F, other: float) -> _F:
		return self.__class__(float(self).__rmod__(other))

	def __rdivmod__(self: _F, other: float) -> Tuple[_F, _F]:
		return tuple(self.__class__(x) for x in float(self).__rdivmod__(other))  # type: ignore

	def __rpow__(self: _F, other: float, mod=None) -> _F:
		return self.__class__(float(self).__rpow__(other, mod))

	def __getnewargs__(self) -> Tuple[float]:
		return self._value

	def __trunc__(self) -> int:
		"""
		Truncates the float to an integer.
		"""

		return float(self).__trunc__()

	def __round__(self, ndigits: Optional[int] = None) -> Union[int, float]:  # type: ignore
		"""
		Round the :class:`~.UserFloat` to ``ndigits`` decimal places, defaulting to ``0``.

		If ``ndigits`` is omitted or :py:obj:`None`, returns an :class:`int`,
		otherwise returns a :class:`float`.
		Rounds half toward even.

		:param ndigits:
		"""

		return float(self).__round__(ndigits)

	def __eq__(self, other: object) -> bool:
		if isinstance(other, UserFloat) and not isinstance(other, float):
			return self._value == other._value
		else:
			return float(self).__eq__(other)

	def __ne__(self, other: object) -> bool:
		if isinstance(other, UserFloat) and not isinstance(other, float):
			return self._value != other._value
		else:
			return float(self).__ne__(other)

	def __lt__(self, other: Union[float, "UserFloat"]) -> bool:
		if isinstance(other, UserFloat) and not isinstance(other, float):
			return self._value < other._value
		else:
			return float(self).__lt__(other)

	def __le__(self, other: Union[float, "UserFloat"]) -> bool:
		if isinstance(other, UserFloat) and not isinstance(other, float):
			return self._value <= other._value
		else:
			return float(self).__le__(other)

	def __gt__(self, other: Union[float, "UserFloat"]) -> bool:
		if isinstance(other, UserFloat) and not isinstance(other, float):
			return self._value > other._value
		else:
			return float(self).__gt__(other)

	def __ge__(self, other: Union[float, "UserFloat"]) -> bool:
		if isinstance(other, UserFloat) and not isinstance(other, float):
			return self._value >= other._value
		else:
			return float(self).__ge__(other)

	def __neg__(self: _F) -> _F:
		return self.__class__(float(self).__neg__())

	def __pos__(self: _F) -> _F:
		return self.__class__(float(self).__pos__())

	def __str__(self) -> str:
		return str(float(self))

	def __int__(self) -> int:
		return int(float(self))

	def __float__(self) -> float:
		return self._value[0]

	def __abs__(self: _F) -> _F:
		return self.__class__(float(self).__abs__())

	def __hash__(self) -> int:
		return float(self).__hash__()

	def __repr__(self) -> str:
		return str(self)

	def __ceil__(self):
		raise NotImplementedError

	def __floor__(self):
		raise NotImplementedError

	def __bool__(self) -> bool:
		"""
		Return ``self != 0``.
		"""

		return super().__bool__()

	def __complex__(self) -> complex:
		"""
		Return :class:`complex(self) <complex>`.

		.. code-block:: python

			complex(self) == complex(float(self), 0)
		"""

		return super().__complex__()


@prettify_docstrings
class NamedList(UserList[_T]):
	"""
	A list with a name.

	The name of the list is taken from the name of the subclass.

	.. versionchanged:: 0.10.0

		:class:`~.NamedList` now subclasses :class:`.UserList` rather than :class:`collections.UserList`.
	"""

	def __repr__(self) -> str:
		return f"{super().__repr__()}"

	def __str__(self) -> str:
		return f"{self.__class__.__name__}{pformat(list(self))}"


def namedlist(name: str = "NamedList") -> Type[NamedList]:
	"""
	A factory function to return a custom list subclass with a name.

	:param name: The name of the list.
	"""

	class cls(NamedList):
		pass

	cls.__name__ = name

	return cls


class Lineup(UserList[_T]):
	"""
	List-like type with fluent methods and some star players.

	.. latex:vspace:: -10px
	"""

	def replace(self: _LU, what: _T, with_: _T) -> _LU:
		r"""
		Replace the first instance of ``what`` with ``with_``.

		:param what: The object to find and replace.
		:param with\_: The new value for the position in the list.
		"""

		self[self.index(what)] = with_

		return self

	def sort(  # type: ignore
			self: _LU,
			*,
			key=None,
			reverse: bool = False,
			) -> _LU:
		"""
		Sort the list in ascending order and return the self.

		The sort is in-place (i.e. the list itself is modified) and stable (i.e. the
		order of two equal elements is maintained).

		If a key function is given, apply it once to each list item and sort them,
		ascending or descending, according to their function values.

		The reverse flag can be set to sort in descending order.
		"""

		super().sort(key=key, reverse=reverse)
		return self

	def reverse(self: _LU) -> _LU:  # type: ignore  # noqa: D102
		super().reverse()
		return self

	def append(  # type: ignore  # noqa: D102
			self: _LU,
			item: _T,
			) -> _LU:
		super().append(item)
		return self

	def extend(  # type: ignore  # noqa: D102
			self: _LU,
			other: Iterable[_T],
			) -> _LU:
		super().extend(other)
		return self

	def insert(  # type: ignore  # noqa: D102
			self: _LU,
			i: int,
			item: _T,
			) -> _LU:
		super().insert(i, item)
		return self

	def remove(  # type: ignore  # noqa: D102
			self: _LU,
			item: _T,
			) -> _LU:
		super().remove(item)
		return self

	def clear(self: _LU) -> _LU:  # type: ignore  # noqa: D102
		super().clear()
		return self



#!/usr/bin/env python
# cython: language_level=3
#
#  delegators.py
"""
Decorators for functions that delegate parts of their functionality to other functions.

.. versionadded:: 0.10.0
"""
#
#  Copyright © 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#
#  delegate_kwargs based on https://github.com/fastai/fastcore
#  |  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  |  not use this file except in compliance with the License. You may obtain
#  |  a copy of the License at
#  |
#  |      http://www.apache.org/licenses/LICENSE-2.0
#  |
#  |  Unless required by applicable law or agreed to in writing, software
#  |  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  |  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  |  License for the specific language governing permissions and limitations
#  |  under the License.
#

# stdlib
import inspect
from typing import Callable, TypeVar, get_type_hints

__all__ = ["delegate_kwargs", "delegates"]

_C = TypeVar("_C", bound="Callable")


def delegate_kwargs(to: Callable, *except_: str) -> Callable[[_C], _C]:
	r"""
	Decorator to replace ``**kwargs`` in function signatures with the
	parameter names from the delegated function.

	:param to: The function \*\*kwargs is passed on to.
	:param \*except\_: Parameter names not to delegate.

	:raises ValueError: if a non-default argument follows a default argument.
	"""  # noqa: D400

	# TODO: return annotation

	def _f(f: _C) -> _C:
		to_f, from_f = to, f

		to_sig = inspect.signature(to_f)
		from_sig = inspect.signature(from_f)
		to_annotations = get_type_hints(to_f)
		from_annotations = get_type_hints(from_f)
		to_params = {k: v for k, v in to_sig.parameters.items() if k not in except_}
		from_params = dict(from_sig.parameters)

		if from_params.pop("kwargs", False):
			if "kwargs" in from_annotations:
				del from_annotations["kwargs"]

			for param in from_params:
				if param in to_params:
					del to_params[param]

			f.__signature__ = from_sig.replace(  # type: ignore
				parameters=[*from_params.values(), *to_params.values()]
				)
			f.__annotations__ = {**to_annotations, **from_annotations}

		return f

	return _f


def delegates(to: Callable) -> Callable[[_C], _C]:
	r"""
	Decorator to replace ``*args, **kwargs``  function signatures
	with the signature of the delegated function.

	:param to: The function the arguments are passed on to.
	"""  # noqa: D400

	def copy_annotations(f):
		if hasattr(to, "__annotations__"):
			if hasattr(f, "__annotations__"):
				return_annotation = f.__annotations__.get("return", inspect.Parameter.empty)
				f.__annotations__.update(to.__annotations__)
				if return_annotation is not inspect.Parameter.empty:
					f.__annotations__["return"] = return_annotation
			else:
				f.__annotations__ = to.__annotations__

	def _f(f: _C) -> _C:
		to_sig = inspect.signature(to)
		from_sig = inspect.signature(f)
		from_params = dict(from_sig.parameters)

		if tuple(from_params.keys()) == ("args", "kwargs"):
			f.__signature__ = to_sig  # type: ignore
			copy_annotations(f)

		elif tuple(from_params.keys()) == ("self", "args", "kwargs"):
			f.__signature__ = from_sig.replace(  # type: ignore
				parameters=[from_params["self"], *to_sig.parameters.values()]
				)

			copy_annotations(f)

		return f

	return _f



# noqa: D100

# stdlib
import sys

if sys.version_info[:2] < (3, 9):  # pragma: no cover (py39+)
	# 3rd party
	import importlib_metadata
	globals().update(importlib_metadata.__dict__)

else:  # pragma: no cover (<py39)
	# stdlib
	import importlib.metadata
	globals().update(importlib.metadata.__dict__)



# From https://github.com/python/typeshed
# Apache-2.0 Licensed

# stdlib
import os
import sys
from pathlib import Path
from types import ModuleType
from typing import Any, BinaryIO, ContextManager, Iterator, TextIO, Union

Package = Union[str, ModuleType]
Resource = Union[str, os.PathLike[Any]]

def open_binary(package: Package, resource: Resource) -> BinaryIO: ...
def open_text(package: Package, resource: Resource, encoding: str = ..., errors: str = ...) -> TextIO: ...
def read_binary(package: Package, resource: Resource) -> bytes: ...
def read_text(package: Package, resource: Resource, encoding: str = ..., errors: str = ...) -> str: ...
def path(package: Package, resource: Resource) -> ContextManager[Path]: ...
def is_resource(package: Package, name: str) -> bool: ...
def contents(package: Package) -> Iterator[str]: ...

if sys.version_info >= (3, 9):
	# stdlib
	from contextlib import AbstractContextManager
	from importlib.abc import Traversable

	def files(package: Package) -> Traversable: ...
	def as_file(path: Traversable) -> AbstractContextManager[Path]: ...



# From https://github.com/dgilland/pydash
# Stripped back to the bare minimum.
#
# MIT License
#
# Copyright (c) 2020 Derrick Gilland
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
# OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
from typing import Iterable, Iterator

__all__ = ["is_match_with", "iterator"]


def is_match_with(obj, source):  # noqa: D103
	if (
			isinstance(obj, dict) and isinstance(source, dict)
			or isinstance(obj, list) and isinstance(source, list)
			or isinstance(obj, tuple) and isinstance(source, tuple)
			):
		# Set equal to True if source is empty, otherwise, False and then allow
		# deep comparison to determine equality.
		equal = not source

		# Walk a/b to determine equality.
		for key, value in iterator(source):
			try:
				equal = is_match_with(obj[key], value)
			except Exception:  # pylint: disable=broad-except
				equal = False

			if not equal:
				break
	else:
		equal = obj == source

	return equal


def iterator(obj) -> Iterator:  # noqa: D103
	if isinstance(obj, dict) or hasattr(obj, "items"):
		return iter(obj.items())
	elif isinstance(obj, Iterable):
		return enumerate(obj)
	else:  # pragma: no cover
		return iter(getattr(obj, "__dict__", {}).items())



#!/usr/bin/env python
# cython: language_level=3
#
#  utils.py
"""
Functions and classes for pretty printing.

.. versionadded:: 0.10.0
"""
#
#  Copyright © 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#
#  Based on CPython.
#  Licensed under the Python Software Foundation License Version 2.
#  Copyright © 2001-2020 Python Software Foundation. All rights reserved.
#  Copyright © 2000 BeOpen.com. All rights reserved.
#  Copyright © 1995-2000 Corporation for National Research Initiatives. All rights reserved.
#  Copyright © 1991-1995 Stichting Mathematisch Centrum. All rights reserved.
#

# stdlib
import sys
from io import StringIO
from typing import IO, Any, Callable, Iterator, MutableMapping, Optional, Tuple, Type, TypeVar

try:  # pragma: no cover

	# 3rd party
	from pprint36 import PrettyPrinter
	from pprint36._pprint import _safe_key  # type: ignore

	supports_sort_dicts = True

except ImportError:

	# stdlib
	from pprint import PrettyPrinter, _safe_key  # type: ignore

	supports_sort_dicts = sys.version_info >= (3, 8)

__all__ = ["FancyPrinter", "simple_repr"]

_T = TypeVar("_T", bound=Type)


class FancyPrinter(PrettyPrinter):
	"""
	Subclass of :class:`~.pprint.PrettyPrinter` with different formatting.

	:param indent: Number of spaces to indent for each level of nesting.
	:param width: Attempted maximum number of columns in the output.
	:param depth: The maximum depth to print out nested structures.
	:param stream: The desired output stream.  If omitted (or :py:obj:`False`), the standard
		output stream available at construction will be used.
	:param compact: If :py:obj:`True`, several items will be combined in one line.
	:param sort_dicts: If :py:obj:`True`, dict keys are sorted.
		Only takes effect on Python 3.8 and later,
		or if `pprint36 <https://pypi.org/project/pprint36/>`_ is installed.
	"""

	def __init__(
			self,
			indent: int = 1,
			width: int = 80,
			depth: Optional[int] = None,
			stream: Optional[IO[str]] = None,
			*,
			compact: bool = False,
			sort_dicts: bool = True,
			):
		if supports_sort_dicts:
			super().__init__(
					indent=indent,
					width=width,
					depth=depth,
					stream=stream,
					compact=compact,
					sort_dicts=sort_dicts,
					)
		else:
			super().__init__(
					indent=indent,
					width=width,
					depth=depth,
					stream=stream,
					compact=compact,
					)

	_dispatch: MutableMapping[Callable, Callable]
	_indent_per_level: int
	_format_items: Callable[[PrettyPrinter, Any, Any, Any, Any, Any, Any], None]
	_dispatch = dict(PrettyPrinter._dispatch)  # type: ignore

	def _make_open(self, char: str, indent: int, obj):
		if self._indent_per_level > 1:
			the_indent = ' ' * (indent + 1)
		else:
			the_indent = ' ' * (indent + self._indent_per_level)

		if obj and not self._compact:  # type: ignore
			return f"{char}\n{the_indent}"
		else:
			return char

	def _make_close(self, char: str, indent: int, obj):
		if obj and not self._compact:  # type: ignore
			return f",\n{' ' * (indent + self._indent_per_level)}{char}"
		else:
			return char

	def _pprint_dict(
			self,
			object,  # noqa: A002  # pylint: disable=redefined-builtin
			stream,
			indent,
			allowance,
			context,
			level,
			):
		obj = object
		write = stream.write
		write(self._make_open('{', indent, obj))

		if self._indent_per_level > 1:
			write((self._indent_per_level - 1) * ' ')

		if obj:
			self._format_dict_items(  # type: ignore
					obj.items(),
					stream,
					indent,
					allowance + 1,
					context,
					level,
					)

		write(self._make_close('}', indent, obj))

	_dispatch[dict.__repr__] = _pprint_dict

	def _pprint_list(self, obj, stream, indent, allowance, context, level):
		stream.write(self._make_open('[', indent, obj))
		self._format_items(obj, stream, indent, allowance + 1, context, level)
		stream.write(self._make_close(']', indent, obj))

	_dispatch[list.__repr__] = _pprint_list

	def _pprint_tuple(self, obj, stream, indent, allowance, context, level):
		stream.write(self._make_open('(', indent, obj))
		endchar = ",)" if len(obj) == 1 else self._make_close(')', indent, obj)
		self._format_items(obj, stream, indent, allowance + len(endchar), context, level)
		stream.write(endchar)

	_dispatch[tuple.__repr__] = _pprint_tuple

	def _pprint_set(self, obj, stream, indent, allowance, context, level):
		if not obj:
			stream.write(repr(obj))
			return

		typ = obj.__class__
		if typ is set:
			stream.write(self._make_open('{', indent, obj))
			endchar = self._make_close('}', indent, obj)
		else:
			stream.write(typ.__name__ + f"({{\n{' ' * (indent + self._indent_per_level + len(typ.__name__) + 1)}")
			endchar = f",\n{' ' * (indent + self._indent_per_level + len(typ.__name__) + 1)}}})"
			indent += len(typ.__name__) + 1

		obj = sorted(obj, key=_safe_key)
		self._format_items(obj, stream, indent, allowance + len(endchar), context, level)
		stream.write(endchar)

	_dispatch[set.__repr__] = _pprint_set
	_dispatch[frozenset.__repr__] = _pprint_set


class Attributes:

	def __init__(self, obj: object, *attributes: str):
		self.obj = obj
		self.attributes = attributes

	def __iter__(self) -> Iterator[Tuple[str, Any]]:
		for attr in self.attributes:
			yield attr, getattr(self.obj, attr)

	def __len__(self) -> int:
		return len(self.attributes)

	def __repr__(self) -> str:
		return f"Attributes{self.attributes}"


class ReprPrettyPrinter(FancyPrinter):

	_dispatch = dict(FancyPrinter._dispatch)

	def format_attributes(self, obj: Attributes):
		stream = StringIO()
		context = {}

		context[id(obj)] = 1

		stream.write(f"(\n{self._indent_per_level * ' '}")

		if self._indent_per_level > 1:
			stream.write((self._indent_per_level - 1) * ' ')

		if obj:
			self._format_attribute_items(list(obj), stream, 0, 0 + 1, context, 1)
		stream.write(f"\n{self._indent_per_level * ' '})")

		del context[id(obj)]
		return stream.getvalue()

	def _format_attribute_items(self, items, stream, indent, allowance, context, level):

		write = stream.write
		indent += self._indent_per_level
		delimnl = ",\n" + ' ' * indent
		last_index = len(items) - 1

		for i, (key, ent) in enumerate(items):
			last = i == last_index
			write(key)
			write('=')
			self._format(  # type: ignore
				ent,
				stream,
				indent + len(key) + 2,
				allowance if last else 1,
				context,
				level,
				)

			if not last:
				write(delimnl)


_default_formatter = ReprPrettyPrinter()


def simple_repr(*attributes: str, show_module: bool = False, **kwargs):
	r"""
	Adds a simple ``__repr__`` method to the decorated class.

	:param attributes: The attributes to include in the ``__repr__``.
	:param show_module: Whether to show the name of the module in the ``__repr__``.
	:param \*\*kwargs: Keyword arguments passed on to :class:`pprint.PrettyPrinter`.
	"""

	def deco(obj: _T) -> _T:

		def __repr__(self) -> str:
			if kwargs:
				formatter = ReprPrettyPrinter(**kwargs)
			else:
				formatter = _default_formatter

			class_name = f"{type(self).__module__}.{type(self).__name__}" if show_module else type(self).__name__

			return f"{class_name}{formatter.format_attributes(Attributes(self, *attributes))}"

		__repr__.__doc__ = f"Return a string representation of the :class:`~{obj.__module__}.{obj.__name__}`."
		__repr__.__name__ = "__repr__"
		__repr__.__module__ = obj.__module__
		__repr__.__qualname__ = f"{obj.__module__}.__repr__"
		obj.__repr__ = __repr__  # type: ignore

		return obj

	return deco
